# -*- coding: utf-8 -*-
"""validate"""


def validate_smearing(parameters, smearing=None):
    """Validate smearing parameters and update the parameters input node accordingly.
    :param parameters: the Dict node that will be used in the inputs
    :param smearing: a tuple of a string and float corresponding to type of smearing and the degauss value
    :raises ValueError: if the input is invalid
    """
    if not any(smearing):
        return

    valid_smearing_types = {
        'gaussian': ['gaussian', 'gauss'],
        'methfessel-paxton': ['methfessel-paxton', 'm-p', 'mp'],
        'marzari-vanderbilt': ['marzari-vanderbilt', 'cold', 'm-v', 'mv'],
        'fermi-dirac': ['fermi-dirac', 'f-d', 'fd'],
    }

    for _, options in valid_smearing_types.items():
        if smearing[0] in options:
            break
    else:
        raise ValueError(
            f"the smearing type \"{smearing[0]}\" is invalid, choose from {', '.join(list(valid_smearing_types.keys()))}"
        )

    if not isinstance(smearing[1], float):
        raise ValueError('the smearing value should be a float')

    parameters['SYSTEM']['occupations'] = 'smearing'
    parameters['SYSTEM']['smearing'] = smearing[0]
    parameters['SYSTEM']['degauss'] = smearing[1]



# -*- coding: utf-8 -*-
# pylint: disable=cyclic-import,reimported,unused-import,wrong-import-position
"""Module with CLI commands for the various work chain implementations."""
from .. import cmd_root


@cmd_root.group('workflow')
def cmd_workflow():
    """Commands to launch and interact with workflows."""


@cmd_workflow.group('launch')
def cmd_launch():
    """Launch workflows."""


from .delta_factor import launch_workflow
from .verification import launch_workflow



# -*- coding: utf-8 -*-
"""Pre-defined overridable options for commonly used command line interface parameters."""
import click

from aiida.cmdline.params.options import OverridableOption

PROTOCOL = OverridableOption('-P',
                             '--protocol',
                             type=click.STRING,
                             default='efficiency',
                             show_default=True,
                             help='The protocol used in verification.')

DUAL = OverridableOption('-D',
                         '--dual',
                         type=click.INT,
                         default=8,
                         show_default=True,
                         help='The dual between ecutwfc and ecutrho.')

MAX_NUM_MACHINES = OverridableOption(
    '-m',
    '--max-num-machines',
    type=click.INT,
    default=1,
    show_default=True,
    help='The maximum number of machines (nodes) to use for the calculations.')

MAX_WALLCLOCK_SECONDS = OverridableOption(
    '-w',
    '--max-wallclock-seconds',
    type=click.INT,
    default=1800,
    show_default=True,
    help='the maximum wallclock time in seconds to set for the calculations.')

WITH_MPI = OverridableOption('-i',
                             '--with-mpi',
                             is_flag=True,
                             default=True,
                             show_default=True,
                             help='Run the calculations with MPI enabled.')

DAEMON = OverridableOption(
    '-d',
    '--daemon',
    is_flag=True,
    default=False,
    show_default=True,
    help='Submit the process to the daemon instead of running it locally.')

AUTOMATIC_PARALLELIZATION = OverridableOption(
    '-a',
    '--automatic-parallelization',
    is_flag=True,
    default=False,
    show_default=True,
    help='Enable the automatic parallelization option of the workchain.')

CLEAN_WORKDIR = OverridableOption(
    '-x',
    '--clean-workdir',
    is_flag=True,
    default=False,
    show_default=True,
    help=
    'Clean the remote folder of all the launched calculations after completion of the workchain.'
)



# -*- coding: utf-8 -*-
# pylint: disable=wrong-import-position,wildcard-import
"""Module for the command line interface."""
import click
import click_completion

from aiida.cmdline.params import types
from aiida.cmdline.params import options as options_core

# Activate the completion of parameter types provided by the click_completion package
click_completion.init()


@click.group('aiida-sssp-workflow',
             context_settings={'help_option_names': ['-h', '--help']})
@options_core.PROFILE(type=types.ProfileParamType(load_profile=True))
def cmd_root(profile):  # pylint: disable=unused-argument
    """CLI for the `aiida-sssp-workflow` plugin."""


from aiida_sssp_workflow.cli.workflows import cmd_workflow



# -*- coding: utf-8 -*-
""" Tests for the plugin.

Includes both tests written in unittest style (test_cli.py) and tests written
in pytest style (test_calculations.py).
"""
import os

TEST_DIR = os.path.dirname(os.path.realpath(__file__))



# -*- coding: utf-8 -*-
"""
Module contain the launch cmdline method of delta-factor workflow
"""
import os
import click

from aiida.plugins import WorkflowFactory, DataFactory
from aiida import orm
from aiida.cmdline.utils import decorators
from aiida.cmdline.params import options as options_core
from aiida.cmdline.params import types
from aiida.cmdline.params.types import PathOrUrl

from . import cmd_launch
from .. import options
from .. import launch


@cmd_launch.command('delta-factor')
@click.argument('pseudo', nargs=1, type=PathOrUrl(exists=True, readable=True))
@options_core.CODE(required=True,
                   type=types.CodeParamType(entry_point='quantumespresso.pw'))
@options.PROTOCOL()
@options.CLEAN_WORKDIR()
@options.MAX_NUM_MACHINES()
@options.MAX_WALLCLOCK_SECONDS()
@options.WITH_MPI()
@options.DAEMON()
@decorators.with_dbenv()
def launch_workflow(code, pseudo, protocol, clean_workdir, max_num_machines,
                    max_wallclock_seconds, with_mpi, daemon):
    """Run the workflow to calculate delta factor"""
    from aiida_sssp_workflow.utils import get_default_options
    UpfData = DataFactory('pseudo.upf')

    builder = WorkflowFactory('sssp_workflow.delta_factor').get_builder()

    pseudo_abspath = os.path.abspath(pseudo)
    with open(pseudo_abspath, 'rb') as stream:
        pseudo = UpfData(stream)

    builder.pseudo = pseudo

    builder.code = code
    builder.protocol = orm.Str(protocol)

    metadata_options = get_default_options(max_num_machines,
                                           max_wallclock_seconds, with_mpi)
    builder.options = orm.Dict(dict=metadata_options)
    builder.clean_workdir = orm.Bool(clean_workdir)

    launch.launch_process(builder, daemon)



# -*- coding: utf-8 -*-
"""
Parsers provided by aiida_sssp_workflow.

Register parsers via the "aiida.parsers" entry point in setup.json.
"""
from aiida.engine import ExitCode
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory

DiffCalculation = CalculationFactory('sssp_workflow')


class DiffParser(Parser):
    """
    Parser class for parsing output of calculation.
    """
    def __init__(self, node):
        """
        Initialize Parser instance

        Checks that the ProcessNode being passed was produced by a DiffCalculation.

        :param node: ProcessNode of calculation
        :param type node: :class:`aiida.orm.ProcessNode`
        """
        from aiida.common import exceptions
        super().__init__(node)
        if not issubclass(node.process_class, DiffCalculation):
            raise exceptions.ParsingError('Can only parse DiffCalculation')

    def parse(self, **kwargs):
        """
        Parse outputs, store results in database.

        :returns: an exit code, if parsing fails (or nothing if parsing succeeds)
        """
        from aiida.orm import SinglefileData

        output_filename = self.node.get_option('output_filename')

        # Check that folder content is as expected
        files_retrieved = self.retrieved.list_object_names()
        files_expected = [output_filename]
        # Note: set(A) <= set(B) checks whether A is a subset of B
        if not set(files_expected) <= set(files_retrieved):
            self.logger.error(
                f"Found files '{files_retrieved}', expected to find '{files_expected}'"
            )
            return self.exit_codes.ERROR_MISSING_OUTPUT_FILES

        # add output file
        self.logger.info(f"Parsing '{output_filename}'")
        with self.retrieved.open(output_filename, 'rb') as handle:
            output_node = SinglefileData(file=handle)
        self.out('sssp_workflow', output_node)

        return ExitCode(0)



# -*- coding: utf-8 -*-
"""
Convergence test on bands of a given pseudopotential
"""
from aiida.common import AttributeDict
from aiida.engine import workfunction
from aiida import orm

from aiida_sssp_workflow.utils import update_dict, NONMETAL_ELEMENTS
from aiida_sssp_workflow.workflows.bands import BandsWorkChain
from aiida_sssp_workflow.calculations.calculate_bands_distance import calculate_bands_distance
from .base import BaseConvergenceWorkChain

PARA_ECUTWFC_LIST = lambda: orm.List(list=[
    20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 110,
    120, 130, 140, 160, 180, 200
])

PARA_ECUTRHO_LIST = lambda: orm.List(list=[
    160, 200, 240, 280, 320, 360, 400, 440, 480, 520, 560, 600, 640, 680, 720,
    760, 800, 880, 960, 1040, 1120, 1280, 1440, 1600
])


@workfunction
def helper_bands_distence_difference(input_band_structure: orm.BandsData,
                                     ref_band_structure: orm.BandsData,
                                     input_band_parameters: orm.Dict,
                                     ref_band_parameters: orm.Dict,
                                     smearing: orm.Float, is_metal: orm.Bool):
    """get the bands distence difference between two calculations"""
    res = calculate_bands_distance(input_band_structure,
                                   ref_band_structure,
                                   input_band_parameters,
                                   ref_band_parameters,
                                   smearing=smearing,
                                   is_metal=is_metal)

    return res


class ConvergenceBandsWorkChain(BaseConvergenceWorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""
    # pylint: disable=too-many-instance-attributes

    _RY_TO_EV = 13.6056980659

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.input('code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        # yapf: enable

    def setup_protocol(self):
        # pylint: disable=invalid-name, attribute-defined-outside-init
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['convergence']['bands_distance']
        self.ctx._DEGAUSS = protocol['degauss']
        self.ctx._OCCUPATIONS = protocol['occupations']
        self.ctx._SMEARING = protocol['smearing']
        self.ctx._CONV_THR_EVA = protocol['electron_conv_thr']
        self.ctx._SCF_KDISTANCE = protocol['scf_kpoints_distance']
        self.ctx._BAND_KDISTANCE = protocol['band_kpoints_distance']
        self.ctx._ININ_NBND_FACTOR = protocol['init_nbnd_factor']

        self.ctx._TOLERANCE = protocol['tolerance']
        self.ctx._CONV_THR_CONV = protocol['convergence_conv_thr']
        self.ctx._CONV_WINDOW = protocol['convergence_window']

    def init_step(self):
        element = self.inputs.pseudo.element
        self.ctx.is_metal = element not in NONMETAL_ELEMENTS

    def get_create_process(self):
        return BandsWorkChain

    def get_evaluate_process(self):
        return helper_bands_distence_difference

    def get_parsed_results(self):
        return {
            'eta_v': ('The difference eta of valence bands', 'eV'),
            'eta_10':
            ('The difference eta of valence bands and conduct bands up to 10.0eV',
             'eV'),
            'max_diff_v': ('The max difference of valence bands', 'eV'),
            'max_diff_10':
            ('The max difference of valence bands and conduct bands up to 10.0eV',
             'eV'),
        }

    def get_converge_y(self):
        return 'max_diff_10', 'eV'

    def get_create_process_inputs(self):
        _PW_PARAS = {   # pylint: disable=invalid-name
            'SYSTEM': {
                'degauss': self.ctx._DEGAUSS,
                'occupations': self.ctx._OCCUPATIONS,
                'smearing': self.ctx._SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self.ctx._CONV_THR_EVA,
            },
        }

        inputs = AttributeDict({
            'code': self.inputs.code,
            'pseudos': self.ctx.pseudos,
            'structure': self.ctx.structure,
            'parameters': {
                'pw':
                orm.Dict(
                    dict=update_dict(_PW_PARAS, self.ctx.base_pw_parameters)),
                'scf_kpoints_distance':
                orm.Float(self.ctx._SCF_KDISTANCE),
                'bands_kpoints_distance':
                orm.Float(self.ctx._BAND_KDISTANCE),
                'nbands_factor':
                orm.Float(self.ctx._ININ_NBND_FACTOR)
            },
        })

        return inputs

    def get_evaluate_process_inputs(self):
        ref_workchain = self.ctx.ref_workchain

        res = {
            'ref_band_parameters': ref_workchain.outputs.band_parameters,
            'ref_band_structure': ref_workchain.outputs.band_structure,
            'smearing': orm.Float(self.ctx._DEGAUSS * self._RY_TO_EV),
            'is_metal': orm.Bool(self.ctx.is_metal),
        }

        return res

    def get_output_input_mapping(self):
        res = orm.Dict(
            dict={
                'band_parameters': 'input_band_parameters',
                'band_structure': 'input_band_structure',
            })
        return res



# -*- coding: utf-8 -*-
"""
All in one verification workchain
"""
from aiida import orm
from aiida.engine import WorkChain

from aiida.plugins import WorkflowFactory, DataFactory

DeltaFactorWorkChain = WorkflowFactory('sssp_workflow.delta_factor')
ConvergenceCohesiveEnergy = WorkflowFactory(
    'sssp_workflow.legacy_convergence.cohesive_energy')
ConvergencePhononFrequencies = WorkflowFactory(
    'sssp_workflow.legacy_convergence.phonon_frequencies')
ConvergencePressureWorkChain = WorkflowFactory(
    'sssp_workflow.legacy_convergence.pressure')
ConvergenceBandsWorkChain = WorkflowFactory(
    'sssp_workflow.legacy_convergence.bands')

UpfData = DataFactory('pseudo.upf')


class VerificationWorkChain(WorkChain):
    """The verification workflow to run all test for the given pseudopotential"""
    _MAX_WALLCLOCK_SECONDS = 1800 * 3

    # ecutwfc evaluate list, the normal reference 200Ry not included
    # since reference will anyway included at final inspect step
    _ECUTWFC_LIST = [
        30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 90, 100, 120, 150
    ]

    @classmethod
    def define(cls, spec):
        super().define(spec)
        # yapf: disable
        spec.input('pw_code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('ph_code', valid_type=orm.Code,
                    help='The `ph.x` code use for the `PhCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, default=lambda: orm.Str('efficiency'),
                    help='The protocol to use for the workchain.')
        spec.input('dual', valid_type=orm.Float,
                    help='The dual to derive ecutrho from ecutwfc.(only for legacy convergence wf).')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options for the `PwCalculations`.')
        spec.input('clean_workdir', valid_type=orm.Bool, default=lambda: orm.Bool(False),
                    help='If `True`, work directories of all called calculation will be cleaned at the end of execution.')

        spec.outline(
            cls.init_setup,
            cls.run_verifications,
            cls.report_and_results,
        )
        spec.output_namespace('delta_factor', dynamic=True,
                            help='results of delta factor calculation.')
        spec.output_namespace('convergence_cohesive_energy', dynamic=True,
                            help='results of convergence cohesive energy calculation.')
        spec.output_namespace('convergence_phonon_frequencies', dynamic=True,
                            help='results of convergence phonon_frequencies calculation.')
        spec.output_namespace('convergence_pressure', dynamic=True,
                            help='results of convergence pressure calculation.')
        spec.output_namespace('convergence_bands', dynamic=True,
                              help='results of convergence bands calculation.')

        spec.exit_code(811, 'WARNING_NOT_ALL_SUB_WORKFLOW_OK',
            message='The sub-workflows {processes} is not finished ok.')
        # yapf: enable

    def init_setup(self):
        """prepare inputs for all verification process"""
        base_inputs = {
            'pseudo': self.inputs.pseudo,
            'protocol': self.inputs.protocol,
            'options': self.inputs.options,
            'parallelization': self.inputs.parallelization,
            'clean_workdir':
            orm.Bool(False),  # not clean for sub-workflow clean at final
        }

        self.ctx.delta_factor_inputs = base_inputs.copy()
        self.ctx.delta_factor_inputs['code'] = self.inputs.pw_code

        self.ctx.phonon_frequencies_inputs = base_inputs.copy()
        self.ctx.phonon_frequencies_inputs['pw_code'] = self.inputs.pw_code
        self.ctx.phonon_frequencies_inputs['ph_code'] = self.inputs.ph_code
        self.ctx.phonon_frequencies_inputs['dual'] = self.inputs.dual

        self.ctx.pressure_inputs = base_inputs.copy()
        self.ctx.pressure_inputs['code'] = self.inputs.pw_code
        self.ctx.pressure_inputs['dual'] = self.inputs.dual

        self.ctx.cohesive_energy_inputs = base_inputs.copy()
        self.ctx.cohesive_energy_inputs['code'] = self.inputs.pw_code
        self.ctx.cohesive_energy_inputs['dual'] = self.inputs.dual

        self.ctx.bands_distance_inputs = base_inputs.copy()
        self.ctx.bands_distance_inputs['code'] = self.inputs.pw_code
        self.ctx.bands_distance_inputs['dual'] = self.inputs.dual

        # to collect workchains in a dict
        self.ctx.workchains = {}

    def run_verifications(self):
        """
        running all verification workflows
        """
        ##
        # delta factor
        ##
        running = self.submit(DeltaFactorWorkChain,
                              **self.ctx.delta_factor_inputs)
        self.report(f'submit workchain delta factor pk={running}')

        self.to_context(verify_delta_factor=running)
        self.ctx.workchains['delta_factor'] = running

        ##
        # phonon frequencies convergence test
        ##
        running = self.submit(ConvergencePhononFrequencies,
                              **self.ctx.phonon_frequencies_inputs)
        self.report(
            f'submit workchain phonon frequencies convergence pk={running.pk}')

        self.to_context(verify_phonon_frequencies=running)
        self.ctx.workchains['convergence_phonon_frequencies'] = running

        ##
        # Cohesive energy
        ##
        running = self.submit(ConvergenceCohesiveEnergy,
                              **self.ctx.cohesive_energy_inputs)
        self.report(
            f'submit workchain cohesive energy convergence pk={running.pk}')

        self.to_context(verify_cohesive_energy=running)
        self.ctx.workchains['convergence_cohesive_energy'] = running

        ##
        # Pressure
        ##
        running = self.submit(ConvergencePressureWorkChain,
                              **self.ctx.pressure_inputs)
        self.report(f'submit workchain pressure convergence pk={running.pk}')

        self.to_context(verify_pressure=running)
        self.ctx.workchains['convergence_pressure'] = running

        # ##
        # # bands
        # ##
        # running = self.submit(ConvergenceBandsWorkChain, **self.ctx.bands_distance_inputs)
        # self.report(
        #     f'submit workchain bands distance convergence pk={running.pk}')

        # self.to_context(verify_bands=running)
        # self.ctx.workchains['convergence_bands_distance'] = running

    def report_and_results(self):
        """result"""
        not_finished_ok = {}
        for wname, workchain in self.ctx.workchains.items():
            if workchain.is_finished_ok:
                for label in workchain.outputs:
                    self.out(f'{wname}.{label}', workchain.outputs[label])
            else:
                self.report(
                    f'The sub-workflow {wname} pk={workchain.pk} not finished ok.'
                )
                not_finished_ok[wname] = workchain.pk

        if not_finished_ok:
            return self.exit_codes.WARNING_NOT_ALL_SUB_WORKFLOW_OK.format(
                processes=not_finished_ok)

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_workdir=True` in the inputs."""
        super().on_terminated()

        if self.inputs.clean_workdir.value is False:
            self.report('remote folders will not be cleaned')
            return

        cleaned_calcs = []

        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, orm.CalcJobNode):
                try:
                    called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                    cleaned_calcs.append(called_descendant.pk)
                except (IOError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )



# -*- coding: utf-8 -*-
"""
Reuse the pure python translate from Nicola's F77 code
Copy and paste from Austin's repository:
https://github.com/zooks97/bieFermi/tree/main/bieFermi/py
"""

import numpy as np
from scipy.special import erfc

FDCUT = 30.0  # Fermi-Dirac cutoff
HMCUT = 10.0  # Hermite cutoff
POSHMA = -0.5634  # Positive Hermite (cold I) `a`


def pyefermi(bands,
             weights,
             nelec: int,
             swidth: float,
             stype: int,
             xacc: float = 1.0e-6,
             jmax: int = 10000,
             nmax: int = 100000) -> float:
    """Find the Fermi energy using bisection."""
    # Get min, max eigenvalue and set as initial bounds
    x1 = np.min(bands)
    x2 = np.max(bands)
    x0 = (x1 + x2) / 2

    # Calculate initial f, fmid
    f = smear(bands, weights, x1, nelec, swidth, stype)
    fmid = smear(bands, weights, x2, nelec, swidth, stype)

    # Find bounds which bracket the Fermi energy
    for n in range(1, nmax):
        if f * fmid >= 0:
            x1 = x0 - n * swidth
            x2 = x0 + (n - 0.5) * swidth
            f = smear(bands, weights, x1, nelec, swidth, stype)
            fmid = smear(bands, weights, x2, nelec, swidth, stype)
        else:
            break
    if f * fmid >= 0:
        raise Exception('Could not bracket Fermi energy. Smearing too small?')

    # Set initial fermi energy guess
    if f < 0.0:
        dx = x2 - x1
        rtb = x1
    else:
        dx = x1 - x2
        rtb = x2

    for _ in range(jmax):
        if np.abs(dx) <= xacc or fmid == 0:
            return rtb
        dx = dx * 0.5
        xmid = rtb + dx
        fmid = smear(bands, weights, xmid, nelec, swidth, stype)
        if fmid <= 0:
            rtb = xmid
    raise Exception('Reached maximum number of bisections.')


def smear(bands, weights, xe: float, nelec: int, swidth: float,
          stype: int) -> float:
    """Calculate smeared value used for bisection."""
    sfuncs = [gaussian, fermid, delthm, spline, poshm, poshm2]

    nkpt, nbnd = bands.shape

    z = 0.0
    for i in range(nkpt):
        for j in range(nbnd):
            x = (xe - bands[i, j]) / swidth
            z += weights[i] * sfuncs[stype - 1](x)
    return z - nelec


def gaussian(x: float) -> float:
    """Gaussian."""
    return 2.0 - erfc(x)


def fermid(x: float) -> float:
    """Fermi-Dirac."""
    x = -x
    if x > FDCUT:
        return 0.0
    elif x < -FDCUT:
        return 2.0
    else:
        return 2.0 / (1.0 + np.exp(x))


def delthm(x: float) -> float:
    """Hermite delta expansion."""
    if x > HMCUT:
        return 2.0
    elif x < -HMCUT:
        return 0.0
    else:
        return (2.0 - erfc(x)) + x * np.exp(-x**2) / np.sqrt(np.pi)


def spline(x: float) -> float:
    """Gaussian spline."""
    x = -x
    if x > 0.0:
        fx = np.sqrt(np.e) / 2 * np.exp(-(x + np.sqrt(2.0) / 2.0)**2)
    else:
        fx = 1.0 - np.sqrt(np.e) / 2 * np.exp(-(x - np.sqrt(2.0 / 2.0))**2)
    return 2.0 * fx


def poshm(x: float) -> float:
    """Positive Hermite (cold I)."""
    if x > HMCUT:
        return 2.0
    elif x < -HMCUT:
        return 0.0
    else:
        return (2.0 -
                erfc(x)) + (-2.0 * POSHMA * x * x + 2.0 * x + POSHMA) * np.exp(
                    -x * x) / np.sqrt(np.pi) / 2.0


def poshm2(x: float) -> float:
    """Positive Hermite (cold II)."""
    if x > HMCUT:
        return 2.0
    elif x < -HMCUT:
        return 0.0
    else:
        return (2.0 - erfc(x - 1.0 / np.sqrt(2.0))) + np.sqrt(2.0) * np.exp(
            -x**2 + np.sqrt(2.0) * x - 0.5) / np.sqrt(np.pi)



# -*- coding: utf-8 -*-
"""fixtures"""
import os
import pytest

pytest_plugins = ['aiida.manage.tests.pytest_fixtures']

# Directory where to store outputs for known inputs (usually tests/data)
DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
STATIC_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                          '_static')


@pytest.fixture(scope='function')
def pp_silicon_sg15():
    """
    Create a aiida-pseudo pp data of sg15 silicon
    """
    from aiida import plugins
    UpfData = plugins.DataFactory('pseudo.upf')

    pp_name = 'Si_ONCV_PBE-1.2.upf'
    pp_path = os.path.join(STATIC_DIR, pp_name)

    with open(pp_path, 'rb') as stream:
        pseudo = UpfData(stream)

    yield pseudo



# -*- coding: utf-8 -*-
"""
Convergence test on cohesive energy of a given pseudopotential
"""
from aiida.engine import calcfunction
from aiida import orm

from aiida_sssp_workflow.utils import update_dict
from aiida_sssp_workflow.workflows.evaluate._phonon_frequencies import PhononFrequenciesWorkChain
from .base import BaseConvergenceWorkChain


@calcfunction
def helper_phonon_frequencies_difference(input_parameters: orm.Dict,
                                         ref_parameters: orm.Dict) -> orm.Dict:
    """
    doc
    """
    import numpy as np

    input_frequencies = input_parameters['dynamical_matrix_0']['frequencies']
    ref_frequencies = ref_parameters['dynamical_matrix_0']['frequencies']
    diffs = np.array(input_frequencies) - np.array(ref_frequencies)
    weights = np.array(ref_frequencies)

    absolute_diff = np.mean(diffs)
    absolute_max_diff = np.amax(diffs)

    relative_diff = np.sqrt(np.mean((diffs / weights)**2)) * 100
    relative_max_diff = np.amax(diffs / weights) * 100

    return orm.Dict(
        dict={
            'relative_diff': relative_diff,
            'relative_max_diff': relative_max_diff,
            'absolute_diff': absolute_diff,
            'absolute_max_diff': absolute_max_diff,
            'absolute_unit': 'cm-1',
            'relative_unit': '%'
        })


PARA_ECUTWFC_LIST = lambda: orm.List(list=[
    20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 110,
    120, 130, 140, 160, 180, 200
])

PARA_ECUTRHO_LIST = lambda: orm.List(list=[
    160, 200, 240, 280, 320, 360, 400, 440, 480, 520, 560, 600, 640, 680, 720,
    760, 800, 880, 960, 1040, 1120, 1280, 1440, 1600
])


class ConvergencePhononFrequenciesWorkChain(BaseConvergenceWorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""
    # pylint: disable=too-many-instance-attributes

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('pw_code',
                   valid_type=orm.Code,
                   help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('ph_code',
                   valid_type=orm.Code,
                   help='The `ph.x` code use for the `PhCalculation`.')

    def setup_protocol(self):
        # pylint: disable=invalid-name, attribute-defined-outside-init
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['convergence']['phonon_frequencies']
        self.ctx._DEGAUSS = protocol['degauss']
        self.ctx._OCCUPATIONS = protocol['occupations']
        self.ctx._SMEARING = protocol['smearing']
        self.ctx._CONV_THR_EVA = protocol['electron_conv_thr']
        self.ctx._QPOINTS_LIST = protocol['qpoints_list']
        self.ctx._KDISTANCE = protocol['kpoints_distance']

        # PH parameters
        self.ctx._TR2_PH = protocol['ph']['tr2_ph']
        self.ctx._EPSILON = protocol['ph']['epsilon']

        self.ctx._TOLERANCE = protocol['tolerance']
        self.ctx._CONV_THR_CONV = protocol['convergence_conv_thr']
        self.ctx._CONV_WINDOW = protocol['convergence_window']

    def get_create_process(self):
        return PhononFrequenciesWorkChain

    def get_evaluate_process(self):
        return helper_phonon_frequencies_difference

    def get_parsed_results(self):
        return {
            'relative_max_diff': ('The relative max phonon difference', '%'),
            'absolute_max_diff':
            ('The absolute max phonon difference', 'cm-1'),
            'relative_diff':
            ('The relative phonon frequencies difference', '%'),
            'absolute_diff':
            ('The absolute phonon frequencies difference', 'cm-1'),
        }

    def get_converge_y(self):
        return 'relative_diff', '%'

    def get_create_process_inputs(self):
        _PW_PARAS = {   # pylint: disable=invalid-name
            'SYSTEM': {
                'degauss': self.ctx._DEGAUSS,
                'occupations': self.ctx._OCCUPATIONS,
                'smearing': self.ctx._SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self.ctx._CONV_THR_EVA,
            },
        }
        _PH_PARAS = {  # pylint: disable=invalid-name
            'INPUTPH': {
                'tr2_ph': self.ctx._TR2_PH,
                'epsil': self.ctx._EPSILON,
            }
        }

        inputs = {
            'pw_code': self.inputs.pw_code,
            'ph_code': self.inputs.ph_code,
            'pseudos': self.ctx.pseudos,
            'structure': self.ctx.structure,
            'parameters': {
                'pw':
                orm.Dict(
                    dict=update_dict(_PW_PARAS, self.ctx.base_pw_parameters)),
                'ph':
                orm.Dict(dict=_PH_PARAS),
                'kpoints_distance':
                orm.Float(self.ctx._KDISTANCE),
                'qpoints':
                orm.List(list=self.ctx._QPOINTS_LIST),
            },
        }

        return inputs

    def get_evaluate_process_inputs(self):
        ref_workchain = self.ctx.ref_workchain

        res = {
            'ref_parameters': ref_workchain.outputs.output_parameters,
        }

        return res

    def get_output_input_mapping(self):
        res = orm.Dict(dict={'output_parameters': 'input_parameters'})
        return res



# -*- coding: utf-8 -*-
"""Workchain to calculate delta factor of specific psp"""
import yaml
import importlib_resources

from aiida import orm
from aiida.engine import WorkChain, ToContext, if_
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import update_dict, \
    MAGNETIC_ELEMENTS, \
    RARE_EARTH_ELEMENTS, \
    helper_parse_upf, get_standard_cif_filename_from_element
from aiida_sssp_workflow.calculations.calculate_delta import calculate_delta
from aiida_sssp_workflow.workflows._eos import _EquationOfStateWorkChain

UpfData = DataFactory('pseudo.upf')


def helper_get_magnetic_inputs(structure: orm.StructureData):
    """
    To set initial magnet to the magnetic system, need to set magnetic order to
    every magnetic element site, with certain pw starting_mainetization parameters.
    """
    MAG_INIT_Mn = {'Mn1': 0.5, 'Mn2': -0.3, 'Mn3': 0.5, 'Mn4': -0.3}  # pylint: disable=invalid-name
    MAG_INIT_O = {'O1': 0.5, 'O2': 0.5, 'O3': -0.5, 'O4': -0.5}  # pylint: disable=invalid-name
    MAG_INIT_Cr = {'Cr1': 0.5, 'Cr2': -0.5}  # pylint: disable=invalid-name

    mag_structure = orm.StructureData(cell=structure.cell, pbc=structure.pbc)
    kind_name = structure.get_kind_names()[0]

    # ferromagnetic
    if kind_name in ['Fe', 'Co', 'Ni']:
        for i, site in enumerate(structure.sites):
            mag_structure.append_atom(position=site.position,
                                      symbols=kind_name)

        parameters = {
            'SYSTEM': {
                'nspin': 2,
                'starting_magnetization': {
                    kind_name: 0.2
                },
            },
        }

    #
    if kind_name in ['Mn', 'O', 'Cr']:
        for i, site in enumerate(structure.sites):
            mag_structure.append_atom(position=site.position,
                                      symbols=kind_name,
                                      name=f'{kind_name}{i+1}')

        if kind_name == 'Mn':
            parameters = {
                'SYSTEM': {
                    'nspin': 2,
                    'starting_magnetization': MAG_INIT_Mn,
                },
            }

        if kind_name == 'O':
            parameters = {
                'SYSTEM': {
                    'nspin': 2,
                    'starting_magnetization': MAG_INIT_O,
                },
            }

        if kind_name == 'Cr':
            parameters = {
                'SYSTEM': {
                    'nspin': 2,
                    'starting_magnetization': MAG_INIT_Cr,
                },
            }

    return mag_structure, parameters


class DeltaFactorWorkChain(WorkChain):
    """Workchain to calculate delta factor of specific pseudopotential"""
    # pylint: disable=too-many-instance-attributes

    _MAX_WALLCLOCK_SECONDS = 1800 * 3

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.input('code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, default=lambda: orm.Str('efficiency'),
                    help='The protocol to use for the workchain.')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options for the `PwCalculations`.')
        spec.input('clean_workdir', valid_type=orm.Bool, default=lambda: orm.Bool(False),
                    help='If `True`, work directories of all called calculation will be cleaned at the end of execution.')
        spec.outline(
            cls.init_setup,
            if_(cls.is_rare_earth_element)(
                cls.extra_setup_for_rare_earth_element,
            ),
            if_(cls.is_magnetic_element)(
                cls.extra_setup_for_magnetic_element,
            ),
            cls.setup_pw_parameters_from_protocol,
            cls.setup_pw_resource_options,
            cls.run_eos,
            cls.inspect_eos,
            cls.calculate_delta_calc,
        )
        spec.expose_outputs(_EquationOfStateWorkChain,
                    namespace_options={'help': 'volume_energy and birch_murnaghan_fit result from EOS.'})
        spec.output('output_delta_factor', valid_type=orm.Dict, required=True,
                    help='The delta factor of the pseudopotential.')
        spec.exit_code(201, 'ERROR_SUB_PROCESS_FAILED_EOS',
                    message=f'The {_EquationOfStateWorkChain.__name__} sub process failed.')
        # yapf: enable

    def _get_protocol(self):
        """Load and read protocol from faml file to a verbose dict"""
        import_path = importlib_resources.path('aiida_sssp_workflow',
                                               'sssp_protocol.yml')
        with import_path as pp_path, open(pp_path, 'rb') as handle:
            self._protocol = yaml.safe_load(handle)  # pylint: disable=attribute-defined-outside-init

            return self._protocol

    def init_setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        element = self.inputs.pseudo.element
        self.ctx.element = element

        self.ctx.pw_parameters = {}
        self.ctx.pseudos = {element: self.inputs.pseudo}

        # Structures for delta factor calculation as provided in
        # http:// molmod.ugent.be/deltacodesdft/
        # Exception for lanthanides use nitride structures from
        # https://doi.org/10.1016/j.commatsci.2014.07.030
        cif_file = get_standard_cif_filename_from_element(element)
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=False)

    def is_rare_earth_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in RARE_EARTH_ELEMENTS

    def extra_setup_for_rare_earth_element(self):
        """Extra setup for rare earth element"""
        import_path = importlib_resources.path('aiida_sssp_workflow.REF.UPFs',
                                               'N.pbe-n-radius_5.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_nitrogen = UpfData(stream)
            self.ctx.pseudos['N'] = upf_nitrogen

        # In rare earth case, increase the initial number of bands,
        # otherwise the occupation will not fill up in the highest band
        # which always trigger the `PwBaseWorkChain` sanity check.
        nbands = self.inputs.pseudo.z_valence + upf_nitrogen.z_valence // 2
        nbands_factor = 2

        extra_parameters = {
            'SYSTEM': {
                'nbnd': int(nbands * nbands_factor),
                'nspin': 2,
                'starting_magnetization': {
                    self.ctx.element: 0.2,
                    'N': 0.0,
                },
            },
        }
        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters,
                                             extra_parameters)

    def is_magnetic_element(self):
        """Check if the element is magnetic"""
        return self.ctx.element in MAGNETIC_ELEMENTS

    def extra_setup_for_magnetic_element(self):
        """Extra setup for magnetic element"""
        # Mn (antiferrimagnetic), O and Cr (antiferromagnetic), Fe, Co, and Ni (ferromagnetic).
        self.ctx.structure, extra_parameters = helper_get_magnetic_inputs(
            self.ctx.structure)
        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters,
                                             extra_parameters)

        # setting pseudos
        pseudos = {}
        pseudo = self.inputs.pseudo
        for kind_name in self.ctx.structure.get_kind_names():
            pseudos[kind_name] = pseudo
        self.ctx.pseudos = pseudos

    def setup_pw_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['delta_factor']
        self._DEGAUSS = protocol['degauss']
        self._OCCUPATIONS = protocol['occupations']
        self._SMEARING = protocol['smearing']
        self._CONV_THR = protocol['electron_conv_thr']

        self._KDISTANCE = protocol['kpoints_distance']
        self._ECUTWFC = protocol['ecutwfc']
        self._SCALE_COUNT = protocol['scale_count']
        self._SCALE_INCREMENT = protocol['scale_increment']

        parameters = {
            'SYSTEM': {
                'degauss': self._DEGAUSS,
                'occupations': self._OCCUPATIONS,
                'smearing': self._SMEARING,
                'ecutwfc': self._ECUTWFC,
            },
            'ELECTRONS': {
                'conv_thr': self._CONV_THR,
            },
        }

        # set the ecutrho according to the type of pseudopotential
        # dual 4 for NC and 8 for all other type of PP.
        upf_header = helper_parse_upf(self.inputs.pseudo)
        if upf_header['pseudo_type'] in ['NC', 'SL']:
            dual = 4.0
        else:
            dual = 8.0
        parameters['SYSTEM']['ecutrho'] = self._ECUTWFC * dual

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters,
                                             parameters)
        self.ctx.kpoints_distance = self._KDISTANCE

        self.report(
            f'The pw parameters for EOS step is: {self.ctx.pw_parameters}')

    def setup_pw_resource_options(self):
        """
        setup resource options and parallelization for `PwCalculation` from inputs
        """
        if 'options' in self.inputs:
            self.ctx.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            self.ctx.options = get_default_options(
                max_wallclock_seconds=self._MAX_WALLCLOCK_SECONDS,
                with_mpi=True)

        if 'parallelization' in self.inputs:
            self.ctx.parallelization = self.inputs.parallelization.get_dict()
        else:
            self.ctx.parallelization = {}

        self.report(f'resource options set to {self.ctx.options}')
        self.report(
            f'parallelization options set to {self.ctx.parallelization}')

    def run_eos(self):
        """run eos workchain"""
        # yapf: disable
        self.report(f'{self.ctx.pw_parameters}')
        inputs = {
            'structure': self.ctx.structure,
            'kpoints_distance': orm.Float(self._KDISTANCE),
            'scale_count': orm.Int(self._SCALE_COUNT),
            'scale_increment': orm.Float(self._SCALE_INCREMENT),
            'metadata': {
                'call_link_label': 'eos'
            },
            'scf': {
                'pw': {
                    'code': self.inputs.code,
                    'pseudos': self.ctx.pseudos,
                    'parameters': orm.Dict(dict=self.ctx.pw_parameters),
                    'metadata': {
                        'options': self.ctx.options
                    },
                    'parallelization': orm.Dict(dict=self.ctx.parallelization),
                },
            }
        }
        # yapf: enable

        running = self.submit(_EquationOfStateWorkChain, **inputs)
        self.report(f'launching _EquationOfStateWorkChain<{running.pk}>')

        return ToContext(workchain_eos=running)

    def inspect_eos(self):
        """Inspect the results of _EquationOfStateWorkChain"""
        workchain = self.ctx.workchain_eos

        if not workchain.is_finished_ok:
            self.report(
                f'_EquationOfStateWorkChain failed with exit status {workchain.exit_status}'
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_EOS

        self.out_many(
            self.exposed_outputs(
                workchain,
                _EquationOfStateWorkChain,
            ))

        self.ctx.output_birch_murnaghan_fit = workchain.outputs.output_birch_murnaghan_fit

    def calculate_delta_calc(self):
        """calculate the delta factor"""
        output_bmf = self.ctx.output_birch_murnaghan_fit

        inputs = {
            'element': orm.Str(self.ctx.element),
            'V0': orm.Float(output_bmf['volume0']),
            'B0': orm.Float(output_bmf['bulk_modulus0']),
            'B1': orm.Float(output_bmf['bulk_deriv0']),
        }
        output_delta_factor = calculate_delta(**inputs)
        self.out('output_delta_factor', output_delta_factor)

        self.report(
            f'The birch murnaghan fitting results are: {output_delta_factor.get_dict()}'
        )

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_workdir=True` in the inputs."""
        super().on_terminated()

        if self.inputs.clean_workdir.value is False:
            self.report('remote folders will not be cleaned')
            return

        cleaned_calcs = []

        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, orm.CalcJobNode):
                try:
                    called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                    cleaned_calcs.append(called_descendant.pk)
                except (IOError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )



# -*- coding: utf-8 -*-
"""
Convergence test on bands of a given pseudopotential
"""
import yaml
import importlib_resources

from aiida import orm
from aiida.engine import WorkChain, if_, append_, ToContext
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import NONMETAL_ELEMENTS, update_dict, \
    RARE_EARTH_ELEMENTS, \
    helper_parse_upf, get_standard_cif_filename_from_element
from aiida_sssp_workflow.workflows.evaluate._bands import BandsWorkChain
from aiida_sssp_workflow.calculations import calculate_bands_distance

UpfData = DataFactory('pseudo.upf')

helper_bands_distence_difference = calculate_bands_distance.calculate_bands_distance


class ConvergenceBandsWorkChain(WorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""
    # pylint: disable=too-many-instance-attributes

    _MAX_WALLCLOCK_SECONDS = 1800 * 3

    # ecutwfc evaluate list, the normal reference 200Ry not included
    # since reference will anyway included at final inspect step
    _ECUTWFC_LIST = [
        30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 90, 100, 120, 150
    ]

    _INIT_NBANDS_FACTOR = 3.0
    _RY_TO_EV = 13.6056980659

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        super().define(spec)
        # yapf: disable
        spec.input('code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, default=lambda: orm.Str('efficiency'),
                    help='The protocol to use for the workchain.')
        spec.input('dual', valid_type=orm.Float,
                    help='The dual to derive ecutrho from ecutwfc.(only for legacy convergence wf).')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options for the `PwCalculations`.')
        spec.input('clean_workdir', valid_type=orm.Bool, default=lambda: orm.Bool(False),
                    help='If `True`, work directories of all called calculation will be cleaned at the end of execution.')

        spec.outline(
            cls.init_setup,
            if_(cls.is_rare_earth_element)(
                cls.extra_setup_for_rare_earth_element, ),
            if_(cls.is_fluorine_element)(
                cls.extra_setup_for_fluorine_element, ),
            cls.setup_code_parameters_from_protocol,
            cls.setup_code_resource_options,
            cls.run_reference,
            cls.run_samples,
            cls.results,
        )

        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include results of all calculations.')

        spec.exit_code(401, 'ERROR_SUB_PROCESS_FAILED',
            message='The sub process for `{label}` did not finish successfully.')
        # yapy: enable

    def _get_protocol(self):
        """Load and read protocol from faml file to a verbose dict"""
        import_path = importlib_resources.path('aiida_sssp_workflow',
                                               'sssp_protocol.yml')
        with import_path as pp_path, open(pp_path, 'rb') as handle:
            self._protocol = yaml.safe_load(handle)  # pylint: disable=attribute-defined-outside-init

            return self._protocol

    def init_setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        self.ctx.pw_parameters = {}
        self.ctx.extra_parameters = {}
        element = self.inputs.pseudo.element
        self.ctx.element = element


        self.ctx.pseudos = {element: self.inputs.pseudo}

        # Structures for convergence verification are all primitive structures
        # the original conventional structure comes from the same CIF files of
        # http:// molmod.ugent.be/deltacodesdft/
        # EXCEPT that for the element fluorine the `SiF4.cif` used for convergence
        # reason. But we do the structure setup for SiF4 in the following step:
        # `cls.extra_setup_for_fluorine_element`
        cif_file = get_standard_cif_filename_from_element(element)
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=True)

        # extra setting for bands convergence
        self.ctx.is_metal = element not in NONMETAL_ELEMENTS

    def is_rare_earth_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in RARE_EARTH_ELEMENTS

    def extra_setup_for_rare_earth_element(self):
        """Extra setup for rare earth element"""
        import_path = importlib_resources.path('aiida_sssp_workflow.REF.UPFs',
                                               'N.pbe-n-radius_5.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_nitrogen = UpfData(stream)
            self.ctx.pseudos['N'] = upf_nitrogen

        # In rare earth case, increase the initial number of bands,
        # otherwise the occupation will not fill up in the highest band
        # which always trigger the `PwBaseWorkChain` sanity check.
        nbands = self.inputs.pseudo.z_valence + upf_nitrogen.z_valence // 2
        nbands_factor = 2

        self.ctx.extra_parameters = {
            'SYSTEM': {
                'nbnd': int(nbands * nbands_factor),
            },
        }

    def is_fluorine_element(self):
        """Check if the element is magnetic"""
        return self.ctx.element == 'F'

    def extra_setup_for_fluorine_element(self):
        """Extra setup for fluorine element"""
        cif_file = get_standard_cif_filename_from_element('SiF4')
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=True)

        # setting pseudos
        import_path = importlib_resources.path(
            'aiida_sssp_workflow.REF.UPFs', 'Si.pbe-n-rrkjus_psl.1.0.0.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_silicon = UpfData(stream)
            self.ctx.pseudos['Si'] = upf_silicon

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['convergence']['bands_distance']
        self._DEGAUSS = protocol['degauss']
        self._OCCUPATIONS = protocol['occupations']
        self._SMEARING = protocol['smearing']
        self._CONV_THR = protocol['electron_conv_thr']
        self._KDISTANCE = protocol['kpoints_distance']

        self._MAX_EVALUATE = protocol['max_evaluate']
        self._REFERENCE_ECUTWFC = protocol['reference_ecutwfc']

        self.ctx.max_evaluate = self._MAX_EVALUATE
        self.ctx.reference_ecutwfc = self._REFERENCE_ECUTWFC

        self.ctx.degauss = self._DEGAUSS    # for band distance calculation

        self.ctx.pw_parameters = {
            'SYSTEM': {
                'degauss': self._DEGAUSS,
                'occupations': self._OCCUPATIONS,
                'smearing': self._SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self._CONV_THR,
            },
        }

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters,
                                        self.ctx.extra_parameters)

        # set the ecutrho according to the type of pseudopotential
        # dual 4 for NC and 8 for all other type of PP.
        upf_header = helper_parse_upf(self.inputs.pseudo)
        if upf_header['pseudo_type'] in ['NC', 'SL']:
            dual = 4.0
        else:
            dual = 8.0

        if 'dual' in self.inputs:
            dual = self.inputs.dual

        self.ctx.dual = dual

        self.ctx.kpoints_distance = self._KDISTANCE

        self.report(
            f'The pw parameters for convergence is: {self.ctx.pw_parameters}'
        )

    def setup_code_resource_options(self):
        """
        setup resource options and parallelization for `PwCalculation` from inputs
        """
        if 'options' in self.inputs:
            self.ctx.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            self.ctx.options = get_default_options(
                max_wallclock_seconds=self._MAX_WALLCLOCK_SECONDS,
                with_mpi=True)

        if 'parallelization' in self.inputs:
            self.ctx.parallelization = self.inputs.parallelization.get_dict()
        else:
            self.ctx.parallelization = {}

        self.report(f'resource options set to {self.ctx.options}')
        self.report(
            f'parallelization options set to {self.ctx.parallelization}')

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation CohesiveWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        # yapf: disable
        inputs = {
            'code': self.inputs.code,
            'pseudos': self.ctx.pseudos,
            'structure': self.ctx.structure,
            'pw_base_parameters': orm.Dict(dict=self.ctx.pw_parameters),
            'ecutwfc': orm.Float(ecutwfc),
            'ecutrho': orm.Float(ecutrho),
            'kpoints_distance': orm.Float(self.ctx.kpoints_distance),
            'init_nbands_factor': orm.Float(self._INIT_NBANDS_FACTOR),
            'options': orm.Dict(dict=self.ctx.options),
            'parallelization': orm.Dict(dict=self.ctx.parallelization),
            'clean_workdir': orm.Bool(False),   # will leave the workdir clean to outer most wf
        }
        # yapf: enable

        return inputs

    def run_reference(self):
        """
        run on reference calculation
        """
        ecutwfc = self.ctx.reference_ecutwfc
        ecutrho = ecutwfc * self.ctx.dual
        inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

        running = self.submit(BandsWorkChain, **inputs)
        self.report(f'launching reference BandsWorkChain<{running.pk}>')

        return ToContext(reference=running)

    def run_samples(self):
        """
        run on all other evaluation sample points
        """
        workchain = self.ctx.reference

        if not workchain.is_finished_ok:
            self.report(
                f'PwBaseWorkChain pk={workchain.pk} for reference run is failed.'
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                label='reference')

        for idx in range(self.ctx.max_evaluate):
            ecutwfc = self._ECUTWFC_LIST[idx]
            ecutrho = ecutwfc * self.ctx.dual
            inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

            running = self.submit(BandsWorkChain, **inputs)
            self.report(f'launching [{idx}] BandsWorkChain<{running.pk}>')

            self.to_context(children=append_(running))

    def results(self):
        """
        results
        """
        reference = self.ctx.reference
        reference_bands_parameters = reference.outputs.output_bands_parameters
        reference_bands_structure = reference.outputs.output_bands_structure

        children = self.ctx.children
        success_children = [
            child for child in children if child.is_finished_ok
        ]

        ecutwfc_list = []
        ecutrho_list = []
        eta_v_list = []
        d_output_parameters = {}

        for child in success_children:
            ecutwfc_list.append(child.inputs.ecutwfc.value)
            ecutrho_list.append(child.inputs.ecutrho.value)

            child_bands_parameters = child.outputs.output_bands_parameters
            child_bands_structure = child.outputs.output_bands_structure
            res = helper_bands_distence_difference(
                bands_structure_a=child_bands_structure,
                bands_parameters_a=child_bands_parameters,
                bands_structure_b=reference_bands_structure,
                bands_parameters_b=reference_bands_parameters,
                smearing=orm.Float(self.ctx.degauss * self._RY_TO_EV),
                is_metal=orm.Bool(self.ctx.is_metal))

            eta_v_list.append(res['eta_v'])

        d_output_parameters['ecutwfc_list'] = ecutwfc_list
        d_output_parameters['ecutrho_list'] = ecutrho_list
        d_output_parameters['eta_v_list'] = eta_v_list

        self.out('output_parameters',
                 orm.Dict(dict=d_output_parameters).store())

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_workdir=True` in the inputs."""
        super().on_terminated()

        if self.inputs.clean_workdir.value is False:
            self.report('remote folders will not be cleaned')
            return

        cleaned_calcs = []

        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, orm.CalcJobNode):
                try:
                    called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                    cleaned_calcs.append(called_descendant.pk)
                except (IOError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )



# -*- coding: utf-8 -*-
"""doc"""


def test_update_dict():
    """test update_dict"""
    from aiida_sssp_workflow.utils import update_dict
    import copy

    old_dict = {
        'SYSTEM': {
            'degauss': 0.02,
            'occupations': 'smearing',
            'smearing': 'marzari-vanderbilt',
        },
        'ELECTRONS': {
            'conv_thr': 1e-10,
        },
    }
    old_expect = copy.deepcopy(old_dict)

    new_inputs = {
        'SYSTEM': {
            'ecutrho': 1600,
            'ecutwfc': 200,
            'smearing': 'fd',
        },
    }

    new_dict = update_dict(old_dict, new_inputs)

    assert new_dict['SYSTEM']['degauss'] == 0.02
    assert new_dict['SYSTEM']['smearing'] == 'fd'
    assert new_dict['SYSTEM']['ecutrho'] == 1600
    assert new_dict['ELECTRONS']['conv_thr'] == 1e-10

    # test that the original one is not modified
    assert old_expect == old_dict


def test_cif_from_element():
    """test cif_from_element"""
    from aiida_sssp_workflow.utils import get_standard_cif_filename_from_element

    fname = get_standard_cif_filename_from_element('Si')
    assert 'Si.cif' in fname

    fname = get_standard_cif_filename_from_element('SiF4')
    assert 'SiF4.cif' in fname

    fname = get_standard_cif_filename_from_element('La')
    assert 'LaN.cif' in fname


def test_to_valid_key():
    """test to_valid_key"""
    from aiida_sssp_workflow.utils import to_valid_key

    assert to_valid_key('Si_ONCV_PBE-1.2') == 'Si_ONCV_PBE_1_2'



# -*- coding: utf-8 -*-
"""
Module contain the launch cmdline method of verification workflow
"""
import os
import click
import numpy as np

from aiida.plugins import WorkflowFactory
from aiida import orm
from aiida.cmdline.utils import decorators
from aiida.cmdline.params.types import PathOrUrl
from aiida.cmdline.params import types
from aiida.cmdline.params.options import OverridableOption

from . import cmd_launch
from .. import launch
from .. import options

PW_CODE = OverridableOption(
    '-X',
    '--pw-code',
    'pw_code',
    type=types.CodeParamType(entry_point='quantumespresso.pw'),
    help='A single code identified by its ID, UUID or label.')

PH_CODE = OverridableOption(
    '-Y',
    '--ph-code',
    'ph_code',
    type=types.CodeParamType(entry_point='quantumespresso.ph'),
    help='A single code identified by its ID, UUID or label.')

REF_ECUTWFC = 200
ecutwfc_list = np.array(
    [30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 90, 100, 120, 150, 200])


@cmd_launch.command('verification')
@click.argument('pseudo', nargs=1, type=PathOrUrl(exists=True, readable=True))
@PW_CODE(required=True)
@PH_CODE(required=True)
@options.PROTOCOL()
@options.DUAL()
@options.CLEAN_WORKDIR()
@options.MAX_NUM_MACHINES()
@options.MAX_WALLCLOCK_SECONDS()
@options.WITH_MPI()
@options.DAEMON()
@decorators.with_dbenv()
def launch_workflow(pw_code, ph_code, pseudo, protocol, dual, clean_workdir,
                    max_num_machines, max_wallclock_seconds, with_mpi, daemon):
    """Run the workflow to calculate delta factor"""
    from aiida_sssp_workflow.utils import get_default_options

    builder = WorkflowFactory('sssp_workflow.verification').get_builder()

    pseudo_abspath = os.path.abspath(pseudo)
    pseudo = orm.UpfData.get_or_create(pseudo_abspath)[0]
    builder.pseudo = pseudo

    builder.pw_code = pw_code
    builder.ph_code = ph_code
    builder.protocol = orm.Str(protocol)

    ecutrho_list = ecutwfc_list * dual
    builder.parameters.ecutwfc_list = orm.List(list=list(ecutwfc_list))
    builder.parameters.ecutrho_list = orm.List(list=list(ecutrho_list))
    builder.parameters.ref_cutoff_pair = orm.List(
        list=[REF_ECUTWFC, REF_ECUTWFC * dual])

    metadata_options = get_default_options(max_num_machines,
                                           max_wallclock_seconds, with_mpi)
    builder.options = orm.Dict(dict=metadata_options)
    builder.clean_workdir = orm.Bool(clean_workdir)

    launch.launch_process(builder, daemon)



# -*- coding: utf-8 -*-
"""
Convergence test on cohesive energy of a given pseudopotential
"""
from aiida.engine import calcfunction
from aiida import orm

from aiida_sssp_workflow.utils import update_dict
from aiida_sssp_workflow.workflows.evaluate._pressure import PressureWorkChain
from .base import BaseConvergenceWorkChain

PARA_ECUTWFC_LIST = lambda: orm.List(list=[
    20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 110,
    120, 130, 140, 160, 180, 200
])

PARA_ECUTRHO_LIST = lambda: orm.List(list=[
    160, 200, 240, 280, 320, 360, 400, 440, 480, 520, 560, 600, 640, 680, 720,
    760, 800, 880, 960, 1040, 1120, 1280, 1440, 1600
])


def helper_get_volume_from_pressure_birch_murnaghan(P, V0, B0, B1):
    """
    Knowing the pressure P and the Birch-Murnaghan equation of state
    parameters, gets the volume the closest to V0 (relatively) that is
    such that P_BirchMurnaghan(V)=P

    retrun unit is (%)
    """
    import numpy as np

    # coefficients of the polynomial in x=(V0/V)^(1/3) (aside from the
    # constant multiplicative factor 3B0/2)
    polynomial = [
        3. / 4. * (B1 - 4.), 0, 1. - 3. / 2. * (B1 - 4.), 0,
        3. / 4. * (B1 - 4.) - 1., 0, 0, 0, 0, -2 * P / (3. * B0)
    ]
    V = min([
        V0 / (x.real**3)
        for x in np.roots(polynomial) if abs(x.imag) < 1e-8 * abs(x.real)
    ],
            key=lambda V: abs(V - V0) / float(V0))

    return abs(V - V0) / V0 * 100


@calcfunction
def helper_pressure_difference(input_parameters: orm.Dict,
                               ref_parameters: orm.Dict, V0: orm.Float,
                               B0: orm.Float, B1: orm.Float) -> orm.Dict:
    """
    doc
    """
    res_pressure = input_parameters['hydrostatic_stress']
    ref_pressure = ref_parameters['hydrostatic_stress']
    absolute_diff = abs(res_pressure - ref_pressure)
    relative_diff = helper_get_volume_from_pressure_birch_murnaghan(
        absolute_diff, V0.value, B0.value, B1.value)

    return orm.Dict(
        dict={
            'relative_diff': relative_diff,
            'absolute_diff': absolute_diff,
            'absolute_unit': 'GPascal',
            'relative_unit': '%'
        })


class ConvergencePressureWorkChain(BaseConvergenceWorkChain):
    """WorkChain to converge test on pressure of input structure"""
    # pylint: disable=too-many-instance-attributes

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('code',
                   valid_type=orm.Code,
                   help='The `pw.x` code use for the `PwCalculation`.')
        spec.input(
            'parameters.v0_b0_b1',
            valid_type=orm.Dict,
            help=
            'birch murnaghan fit results used in residual volume evaluation.')

    def setup_protocol(self):
        # pylint: disable=invalid-name, attribute-defined-outside-init
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['convergence']['pressure']
        self.ctx._DEGAUSS = protocol['degauss']
        self.ctx._OCCUPATIONS = protocol['occupations']
        self.ctx._SMEARING = protocol['smearing']
        self.ctx._CONV_THR_EVA = protocol['electron_conv_thr']
        self.ctx._KDISTANCE = protocol['kpoints_distance']

        self.ctx._TOLERANCE = protocol['tolerance']
        self.ctx._CONV_THR_CONV = protocol['convergence_conv_thr']
        self.ctx._CONV_WINDOW = protocol['convergence_window']

    def get_create_process(self):
        return PressureWorkChain

    def get_evaluate_process(self):
        return helper_pressure_difference

    def get_parsed_results(self):
        return {
            'absolute_diff':
            ('The absolute residual pressure difference', 'GPascal'),
            'relative_diff':
            ('The relative residual pressure difference', '%'),
        }

    def get_converge_y(self):
        return 'relative_diff', '%'

    def get_create_process_inputs(self):
        _PW_PARAS = {   # pylint: disable=invalid-name
            'SYSTEM': {
                'degauss': self.ctx._DEGAUSS,
                'occupations': self.ctx._OCCUPATIONS,
                'smearing': self.ctx._SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self.ctx._CONV_THR_EVA,
            },
        }

        inputs = {
            'code': self.inputs.code,
            'pseudos': self.ctx.pseudos,
            'structure': self.ctx.structure,
            'parameters': {
                'pw':
                orm.Dict(
                    dict=update_dict(_PW_PARAS, self.ctx.base_pw_parameters)),
                'kpoints_distance':
                orm.Float(self.ctx._KDISTANCE),
            },
        }

        return inputs

    def get_evaluate_process_inputs(self):
        ref_workchain = self.ctx.ref_workchain
        V0_B0_B1 = self.inputs.parameters.v0_b0_b1  # pylint: disable=invalid-name

        res = {
            'ref_parameters': ref_workchain.outputs.output_parameters,
            'V0': orm.Float(V0_B0_B1['V0']),
            'B0': orm.Float(V0_B0_B1['B0']),
            'B1': orm.Float(V0_B0_B1['B1']),
        }

        return res

    def get_output_input_mapping(self):
        res = orm.Dict(dict={'output_parameters': 'input_parameters'})
        return res



# -*- coding: utf-8 -*-
"""
Convergence test on phonon frequencies of a given pseudopotential
"""
import yaml
import importlib_resources

from aiida.engine import calcfunction
from aiida import orm
from aiida.engine import WorkChain, if_, append_, ToContext
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import update_dict, \
    RARE_EARTH_ELEMENTS, \
    helper_parse_upf, get_standard_cif_filename_from_element
from aiida_sssp_workflow.workflows.evaluate._phonon_frequencies import PhononFrequenciesWorkChain

UpfData = DataFactory('pseudo.upf')


@calcfunction
def helper_phonon_frequencies_difference(input_parameters: orm.Dict,
                                         ref_parameters: orm.Dict) -> orm.Dict:
    """
    doc
    """
    import numpy as np

    input_frequencies = input_parameters['dynamical_matrix_0']['frequencies']
    ref_frequencies = ref_parameters['dynamical_matrix_0']['frequencies']
    diffs = np.array(input_frequencies) - np.array(ref_frequencies)
    weights = np.array(ref_frequencies)

    absolute_diff = np.mean(diffs)
    absolute_max_diff = np.amax(diffs)

    relative_diff = np.sqrt(np.mean((diffs / weights)**2)) * 100
    relative_max_diff = np.amax(diffs / weights) * 100

    return orm.Dict(
        dict={
            'relative_diff': relative_diff,
            'relative_max_diff': relative_max_diff,
            'absolute_diff': absolute_diff,
            'absolute_max_diff': absolute_max_diff,
            'absolute_unit': 'cm-1',
            'relative_unit': '%'
        })


class ConvergencePhononFrequenciesWorkChain(WorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""
    # pylint: disable=too-many-instance-attributes

    _MAX_WALLCLOCK_SECONDS = 1800 * 3

    # ecutwfc evaluate list, the normal reference 200Ry not included
    # since reference will anyway included at final inspect step
    _ECUTWFC_LIST = [
        30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 90, 100, 120, 150
    ]

    @classmethod
    def define(cls, spec):
        super().define(spec)
        # yapf: disable
        spec.input('pw_code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('ph_code', valid_type=orm.Code,
                    help='The `ph.x` code use for the `PhCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, default=lambda: orm.Str('efficiency'),
                    help='The protocol to use for the workchain.')
        spec.input('dual', valid_type=orm.Float,
                    help='The dual to derive ecutrho from ecutwfc.(only for legacy convergence wf).')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options for the `PwCalculations`.')
        spec.input('clean_workdir', valid_type=orm.Bool, default=lambda: orm.Bool(False),
                    help='If `True`, work directories of all called calculation will be cleaned at the end of execution.')

        spec.outline(
            cls.init_setup,
            if_(cls.is_rare_earth_element)(
                cls.extra_setup_for_rare_earth_element, ),
            if_(cls.is_fluorine_element)(
                cls.extra_setup_for_fluorine_element, ),
            cls.setup_code_parameters_from_protocol,
            cls.setup_code_resource_options,
            cls.run_reference,
            cls.run_samples,
            cls.results,
        )

        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include results of all calculations.')

        spec.exit_code(401, 'ERROR_SUB_PROCESS_FAILED',
            message='The sub process for `{label}` did not finish successfully.')
        # yapy: enable

    def _get_protocol(self):
        """Load and read protocol from faml file to a verbose dict"""
        import_path = importlib_resources.path('aiida_sssp_workflow',
                                               'sssp_protocol.yml')
        with import_path as pp_path, open(pp_path, 'rb') as handle:
            self._protocol = yaml.safe_load(handle)  # pylint: disable=attribute-defined-outside-init

            return self._protocol

    def init_setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        element = self.inputs.pseudo.element
        self.ctx.element = element

        self.ctx.pseudos = {element: self.inputs.pseudo}

        # Structures for convergence verification are all primitive structures
        # the original conventional structure comes from the same CIF files of
        # http:// molmod.ugent.be/deltacodesdft/
        # EXCEPT that for the element fluorine the `SiF4.cif` used for convergence
        # reason. But we do the structure setup for SiF4 in the following step:
        # `cls.extra_setup_for_fluorine_element`
        cif_file = get_standard_cif_filename_from_element(element)
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=True)

    def is_rare_earth_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in RARE_EARTH_ELEMENTS

    def extra_setup_for_rare_earth_element(self):
        """Extra setup for rare earth element"""
        import_path = importlib_resources.path('aiida_sssp_workflow.REF.UPFs',
                                               'N.pbe-n-radius_5.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_nitrogen = UpfData(stream)
            self.ctx.pseudos['N'] = upf_nitrogen

        # In rare earth case, increase the initial number of bands,
        # otherwise the occupation will not fill up in the highest band
        # which always trigger the `PwBaseWorkChain` sanity check.
        nbands = self.inputs.pseudo.z_valence + upf_nitrogen.z_valence // 2
        nbands_factor = 2

        extra_parameters = {
            'SYSTEM': {
                'nbnd': int(nbands * nbands_factor),
            },
        }
        self.ctx.bulk_parameters = update_dict(self.ctx.pw_parameters,
                                               extra_parameters)

    def is_fluorine_element(self):
        """Check if the element is magnetic"""
        return self.ctx.element == 'F'

    def extra_setup_for_fluorine_element(self):
        """Extra setup for fluorine element"""
        cif_file = get_standard_cif_filename_from_element('SiF4')
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=True)

        # setting pseudos
        import_path = importlib_resources.path(
            'aiida_sssp_workflow.REF.UPFs', 'Si.pbe-n-rrkjus_psl.1.0.0.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_silicon = UpfData(stream)
            self.ctx.pseudos['Si'] = upf_silicon

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['convergence']['phonon_frequencies']
        self._DEGAUSS = protocol['degauss']
        self._OCCUPATIONS = protocol['occupations']
        self._SMEARING = protocol['smearing']
        self._CONV_THR = protocol['electron_conv_thr']
        self._KDISTANCE = protocol['kpoints_distance']
        self._QPOINTS_LIST = protocol['qpoints_list']

        self._PH_EPSILON = protocol['ph']['epsilon']
        self._PH_TR2_PH = protocol['ph']['tr2_ph']

        self._MAX_EVALUATE = protocol['max_evaluate']
        self._REFERENCE_ECUTWFC = protocol['reference_ecutwfc']

        self.ctx.qpoints_list = self._QPOINTS_LIST

        self.ctx.max_evaluate = self._MAX_EVALUATE
        self.ctx.reference_ecutwfc = self._REFERENCE_ECUTWFC

        self.ctx.pw_parameters = {
            'SYSTEM': {
                'degauss': self._DEGAUSS,
                'occupations': self._OCCUPATIONS,
                'smearing': self._SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self._CONV_THR,
            },
            'CONTROL': {
                'calculation': 'scf',
                'wf_collect': True,
                'tstress': True,
            },
        }

        self.ctx.ph_parameters = {
            'INPUTPH': {
                'tr2_ph': self._PH_TR2_PH,
                'epsil': self._PH_EPSILON,
            }
        }

        # set the ecutrho according to the type of pseudopotential
        # dual 4 for NC and 8 for all other type of PP.
        upf_header = helper_parse_upf(self.inputs.pseudo)
        if upf_header['pseudo_type'] in ['NC', 'SL']:
            dual = 4.0
        else:
            dual = 8.0

        if 'dual' in self.inputs:
            dual = self.inputs.dual

        self.ctx.dual = dual

        self.ctx.kpoints_distance = self._KDISTANCE

        self.report(
            f'The pw parameters for convergence is: {self.ctx.pw_parameters}'
        )
        self.report(
            f'The ph parameters for convergence is: {self.ctx.ph_parameters}'
        )

    def setup_code_resource_options(self):
        """
        setup resource options and parallelization for `PwCalculation` from inputs
        """
        if 'options' in self.inputs:
            self.ctx.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            self.ctx.options = get_default_options(
                max_wallclock_seconds=self._MAX_WALLCLOCK_SECONDS,
                with_mpi=True)

        if 'parallelization' in self.inputs:
            self.ctx.parallelization = self.inputs.parallelization.get_dict()
        else:
            self.ctx.parallelization = {}

        self.report(f'resource options set to {self.ctx.options}')
        self.report(
            f'parallelization options set to {self.ctx.parallelization}')

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation CohesiveWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        # yapf: disable
        inputs = {
            'pw_code': self.inputs.pw_code,
            'ph_code': self.inputs.ph_code,
            'pseudos': self.ctx.pseudos,
            'structure': self.ctx.structure,
            'pw_base_parameters': orm.Dict(dict=self.ctx.pw_parameters),
            'ph_base_parameters': orm.Dict(dict=self.ctx.ph_parameters),
            'ecutwfc': orm.Float(ecutwfc),
            'ecutrho': orm.Float(ecutrho),
            'qpoints': orm.List(list=self.ctx.qpoints_list),
            'kpoints_distance': orm.Float(self.ctx.kpoints_distance),
            'options': orm.Dict(dict=self.ctx.options),
            'parallelization': orm.Dict(dict=self.ctx.parallelization),
            'clean_workdir': orm.Bool(False),   # will leave the workdir clean to outer most wf
        }
        # yapf: enable

        return inputs

    def run_reference(self):
        """
        run on reference calculation
        """
        ecutwfc = self.ctx.reference_ecutwfc
        ecutrho = ecutwfc * self.ctx.dual
        inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

        running = self.submit(PhononFrequenciesWorkChain, **inputs)
        self.report(
            f'launching reference PhononFrequenciesWorkChain<{running.pk}>')

        return ToContext(reference=running)

    def run_samples(self):
        """
        run on all other evaluation sample points
        """
        workchain = self.ctx.reference

        if not workchain.is_finished_ok:
            self.report(
                f'PwBaseWorkChain pk={workchain.pk} for reference run is failed.'
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                label='reference')

        for idx in range(self.ctx.max_evaluate):
            ecutwfc = self._ECUTWFC_LIST[idx]
            ecutrho = ecutwfc * self.ctx.dual
            inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

            running = self.submit(PhononFrequenciesWorkChain, **inputs)
            self.report(
                f'launching [{idx}] PhononFrequenciesWorkChain<{running.pk}>')

            self.to_context(children=append_(running))

    def results(self):
        """
        results
        """
        reference = self.ctx.reference
        reference_parameters = reference.outputs.output_parameters

        children = self.ctx.children
        success_children = [
            child for child in children if child.is_finished_ok
        ]

        ecutwfc_list = []
        ecutrho_list = []
        absolute_diff_list = []
        d_output_parameters = {}

        for child in success_children:
            ecutwfc_list.append(child.inputs.ecutwfc.value)
            ecutrho_list.append(child.inputs.ecutrho.value)

            child_parameters = child.outputs.output_parameters
            res = helper_phonon_frequencies_difference(child_parameters,
                                                       reference_parameters)

            absolute_diff_list.append(res['absolute_diff'])

        d_output_parameters['ecutwfc_list'] = ecutwfc_list
        d_output_parameters['ecutrho_list'] = ecutrho_list
        d_output_parameters['absolute_list'] = absolute_diff_list

        self.out('output_parameters',
                 orm.Dict(dict=d_output_parameters).store())

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_workdir=True` in the inputs."""
        super().on_terminated()

        if self.inputs.clean_workdir.value is False:
            self.report('remote folders will not be cleaned')
            return

        cleaned_calcs = []

        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, orm.CalcJobNode):
                try:
                    called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                    cleaned_calcs.append(called_descendant.pk)
                except (IOError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )



# -*- coding: utf-8 -*-
"""
Convergence test on pressure of a given pseudopotential
"""
import yaml
import importlib_resources

from aiida.engine import calcfunction
from aiida import orm
from aiida.engine import WorkChain, if_, append_, ToContext
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import update_dict, \
    RARE_EARTH_ELEMENTS, \
    helper_parse_upf, get_standard_cif_filename_from_element
from aiida_sssp_workflow.workflows.evaluate._pressure import PressureWorkChain
from aiida_sssp_workflow.workflows._eos import _EquationOfStateWorkChain

UpfData = DataFactory('pseudo.upf')


def helper_get_volume_from_pressure_birch_murnaghan(P, V0, B0, B1):
    """
    Knowing the pressure P and the Birch-Murnaghan equation of state
    parameters, gets the volume the closest to V0 (relatively) that is
    such that P_BirchMurnaghan(V)=P

    retrun unit is (%)
    """
    import numpy as np

    # coefficients of the polynomial in x=(V0/V)^(1/3) (aside from the
    # constant multiplicative factor 3B0/2)
    polynomial = [
        3. / 4. * (B1 - 4.), 0, 1. - 3. / 2. * (B1 - 4.), 0,
        3. / 4. * (B1 - 4.) - 1., 0, 0, 0, 0, -2 * P / (3. * B0)
    ]
    V = min([
        V0 / (x.real**3)
        for x in np.roots(polynomial) if abs(x.imag) < 1e-8 * abs(x.real)
    ],
            key=lambda V: abs(V - V0) / float(V0))

    return abs(V - V0) / V0 * 100


@calcfunction
def helper_pressure_difference(input_parameters: orm.Dict,
                               ref_parameters: orm.Dict, V0: orm.Float,
                               B0: orm.Float, B1: orm.Float) -> orm.Dict:
    """
    doc
    """
    res_pressure = input_parameters['hydrostatic_stress']
    ref_pressure = ref_parameters['hydrostatic_stress']
    absolute_diff = abs(res_pressure - ref_pressure)
    relative_diff = helper_get_volume_from_pressure_birch_murnaghan(
        absolute_diff, V0.value, B0.value, B1.value)

    return orm.Dict(
        dict={
            'relative_diff': relative_diff,
            'absolute_diff': absolute_diff,
            'absolute_unit': 'GPascal',
            'relative_unit': '%'
        })


class ConvergencePressureWorkChain(WorkChain):
    """WorkChain to converge test on pressure of input structure"""
    # pylint: disable=too-many-instance-attributes

    _MAX_WALLCLOCK_SECONDS = 1800 * 3

    # ecutwfc evaluate list, the normal reference 200Ry not included
    # since reference will anyway included at final inspect step
    _ECUTWFC_LIST = [
        30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 90, 100, 120, 150
    ]

    # parameters control inner EOS reference workflow
    _EOS_SCALE_COUNT = 7
    _EOS_SCALE_INCREMENT = 0.02

    @classmethod
    def define(cls, spec):
        super().define(spec)
        # yapf: disable
        spec.input('code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, default=lambda: orm.Str('efficiency'),
                    help='The protocol to use for the workchain.')
        spec.input('dual', valid_type=orm.Float,
                    help='The dual to derive ecutrho from ecutwfc.(only for legacy convergence wf).')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options for the `PwCalculations`.')
        spec.input('clean_workdir', valid_type=orm.Bool, default=lambda: orm.Bool(False),
                    help='If `True`, work directories of all called calculation will be cleaned at the end of execution.')

        spec.outline(
            cls.init_setup,
            if_(cls.is_rare_earth_element)(
                cls.extra_setup_for_rare_earth_element, ),
            if_(cls.is_fluorine_element)(
                cls.extra_setup_for_fluorine_element, ),
            cls.setup_code_parameters_from_protocol,
            cls.setup_code_resource_options,
            cls.run_reference,
            cls.run_extra_reference,
            cls.run_samples,
            cls.results,
        )

        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include results of all calculations.')

        spec.exit_code(401, 'ERROR_SUB_PROCESS_FAILED',
            message='The sub process for `{label}` did not finish successfully.')
        # yapy: enable

    def _get_protocol(self):
        """Load and read protocol from faml file to a verbose dict"""
        import_path = importlib_resources.path('aiida_sssp_workflow',
                                               'sssp_protocol.yml')
        with import_path as pp_path, open(pp_path, 'rb') as handle:
            self._protocol = yaml.safe_load(handle)  # pylint: disable=attribute-defined-outside-init

            return self._protocol

    def init_setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        self.ctx.pw_parameters = {}
        self.ctx.extra_parameters = {}
        element = self.inputs.pseudo.element
        self.ctx.element = element

        self.ctx.pseudos = {element: self.inputs.pseudo}

        # Structures for convergence verification are all primitive structures
        # the original conventional structure comes from the same CIF files of
        # http:// molmod.ugent.be/deltacodesdft/
        # EXCEPT that for the element fluorine the `SiF4.cif` used for convergence
        # reason. But we do the structure setup for SiF4 in the following step:
        # `cls.extra_setup_for_fluorine_element`
        cif_file = get_standard_cif_filename_from_element(element)
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=True)

    def is_rare_earth_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in RARE_EARTH_ELEMENTS

    def extra_setup_for_rare_earth_element(self):
        """Extra setup for rare earth element"""
        import_path = importlib_resources.path('aiida_sssp_workflow.REF.UPFs',
                                               'N.pbe-n-radius_5.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_nitrogen = UpfData(stream)
            self.ctx.pseudos['N'] = upf_nitrogen

        # In rare earth case, increase the initial number of bands,
        # otherwise the occupation will not fill up in the highest band
        # which always trigger the `PwBaseWorkChain` sanity check.
        nbands = self.inputs.pseudo.z_valence + upf_nitrogen.z_valence // 2
        nbands_factor = 2

        self.ctx.extra_parameters = {
            'SYSTEM': {
                'nbnd': int(nbands * nbands_factor),
            },
        }

    def is_fluorine_element(self):
        """Check if the element is magnetic"""
        return self.ctx.element == 'F'

    def extra_setup_for_fluorine_element(self):
        """Extra setup for fluorine element"""
        cif_file = get_standard_cif_filename_from_element('SiF4')
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=True)

        # setting pseudos
        import_path = importlib_resources.path(
            'aiida_sssp_workflow.REF.UPFs', 'Si.pbe-n-rrkjus_psl.1.0.0.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_silicon = UpfData(stream)
            self.ctx.pseudos['Si'] = upf_silicon


    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['convergence']['pressure']
        self._DEGAUSS = protocol['degauss']
        self._OCCUPATIONS = protocol['occupations']
        self._SMEARING = protocol['smearing']
        self._CONV_THR = protocol['electron_conv_thr']
        self._KDISTANCE = protocol['kpoints_distance']

        self._MAX_EVALUATE = protocol['max_evaluate']
        self._REFERENCE_ECUTWFC = protocol['reference_ecutwfc']

        self.ctx.max_evaluate = self._MAX_EVALUATE
        self.ctx.reference_ecutwfc = self._REFERENCE_ECUTWFC

        self.ctx.pw_parameters = {
            'SYSTEM': {
                'degauss': self._DEGAUSS,
                'occupations': self._OCCUPATIONS,
                'smearing': self._SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self._CONV_THR,
            },
            'CONTROL': {
                'calculation': 'scf',
                'wf_collect': True,
                'tstress': True,
            },
        }

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters,
                                        self.ctx.extra_parameters)

        # set the ecutrho according to the type of pseudopotential
        # dual 4 for NC and 8 for all other type of PP.
        upf_header = helper_parse_upf(self.inputs.pseudo)
        if upf_header['pseudo_type'] in ['NC', 'SL']:
            dual = 4.0
        else:
            dual = 8.0

        if 'dual' in self.inputs:
            dual = self.inputs.dual

        self.ctx.dual = dual

        self.ctx.kpoints_distance = self._KDISTANCE

        self.report(
            f'The pw parameters for convergence is: {self.ctx.pw_parameters}'
        )

    def setup_code_resource_options(self):
        """
        setup resource options and parallelization for `PwCalculation` from inputs
        """
        if 'options' in self.inputs:
            self.ctx.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            self.ctx.options = get_default_options(
                max_wallclock_seconds=self._MAX_WALLCLOCK_SECONDS,
                with_mpi=True)

        if 'parallelization' in self.inputs:
            self.ctx.parallelization = self.inputs.parallelization.get_dict()
        else:
            self.ctx.parallelization = {}

        self.report(f'resource options set to {self.ctx.options}')
        self.report(
            f'parallelization options set to {self.ctx.parallelization}')

    def _get_inputs(self, ecutwfc, ecutrho):
        """
        get inputs for the evaluation CohesiveWorkChain by provide ecutwfc and ecutrho,
        all other parameters are fixed for the following steps
        """
        # yapf: disable
        inputs = {
            'code': self.inputs.code,
            'pseudos': self.ctx.pseudos,
            'structure': self.ctx.structure,
            'pw_base_parameters': orm.Dict(dict=self.ctx.pw_parameters),
            'ecutwfc': orm.Float(ecutwfc),
            'ecutrho': orm.Float(ecutrho),
            'kpoints_distance': orm.Float(self.ctx.kpoints_distance),
            'options': orm.Dict(dict=self.ctx.options),
            'parallelization': orm.Dict(dict=self.ctx.parallelization),
            'clean_workdir': orm.Bool(False),   # will leave the workdir clean to outer most wf
        }
        # yapf: enable

        return inputs

    def run_reference(self):
        """
        run on reference calculation
        """
        ecutwfc = self.ctx.reference_ecutwfc
        ecutrho = ecutwfc * self.ctx.dual
        inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

        running = self.submit(PressureWorkChain, **inputs)
        self.report(f'launching reference PressureWorkChain<{running.pk}>')

        return ToContext(reference=running)

    def run_extra_reference(self):
        """
        For pressure convergence workflow, the birch murnagen fitting result is used to
        calculating the pressure. There is an extra workflow (run at ecutwfc of reference point)
        for it which need to be run before the following step.

        This workflow is shared with delta factor workchain for birch murnagan fitting.
        """
        ecutwfc = self.ctx.reference_ecutwfc
        ecutrho = ecutwfc * self.ctx.dual
        parameters = {
            'SYSTEM': {
                'ecutwfc': ecutwfc,
                'ecutrho': ecutrho,
            },
        }
        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters,
                                             parameters)

        self.report(f'{self.ctx.pw_parameters}')
        # yapf: disable
        inputs = {
            'structure': self.ctx.structure,
            'kpoints_distance': orm.Float(self._KDISTANCE),
            'scale_count': orm.Int(self._EOS_SCALE_COUNT),
            'scale_increment': orm.Float(self._EOS_SCALE_INCREMENT),
            'metadata': {
                'call_link_label': 'EOS'
            },
            'scf': {
                'pw': {
                    'code': self.inputs.code,
                    'pseudos': self.ctx.pseudos,
                    'parameters': orm.Dict(dict=self.ctx.pw_parameters),
                    'metadata': {
                        'options': self.ctx.options
                    },
                    'parallelization': orm.Dict(dict=self.ctx.parallelization),
                },
            }
        }
        # yapf: enable

        running = self.submit(_EquationOfStateWorkChain, **inputs)
        self.report(f'launching _EquationOfStateWorkChain<{running.pk}>')

        return ToContext(extra_reference=running)

    def run_samples(self):
        """
        run on all other evaluation sample points
        """
        workchain = self.ctx.reference

        if not workchain.is_finished_ok:
            self.report(
                f'PwBaseWorkChain pk={workchain.pk} for reference run is failed.'
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                label='reference')

        for idx in range(self.ctx.max_evaluate):
            ecutwfc = self._ECUTWFC_LIST[idx]
            ecutrho = ecutwfc * self.ctx.dual
            inputs = self._get_inputs(ecutwfc=ecutwfc, ecutrho=ecutrho)

            running = self.submit(PressureWorkChain, **inputs)
            self.report(f'launching [{idx}] PressureWorkChain<{running.pk}>')

            self.to_context(children=append_(running))

    def results(self):
        """
        results
        """
        reference = self.ctx.reference
        reference_parameters = reference.outputs.output_parameters

        extra_reference = self.ctx.extra_reference
        extra_reference_parameters = extra_reference.outputs.output_birch_murnaghan_fit

        V0 = extra_reference_parameters['volume0']
        B0 = extra_reference_parameters['bulk_modulus0']
        B1 = extra_reference_parameters['bulk_deriv0']

        children = self.ctx.children
        success_children = [
            child for child in children if child.is_finished_ok
        ]

        ecutwfc_list = []
        ecutrho_list = []
        absolute_diff_list = []
        d_output_parameters = {}

        for child in success_children:
            ecutwfc_list.append(child.inputs.ecutwfc.value)
            ecutrho_list.append(child.inputs.ecutrho.value)

            child_parameters = child.outputs.output_parameters
            res = helper_pressure_difference(child_parameters,
                                             reference_parameters,
                                             V0=orm.Float(V0),
                                             B0=orm.Float(B0),
                                             B1=orm.Float(B1))

            absolute_diff_list.append(res['absolute_diff'])

        d_output_parameters['ecutwfc_list'] = ecutwfc_list
        d_output_parameters['ecutrho_list'] = ecutrho_list
        d_output_parameters['absolute_list'] = absolute_diff_list

        self.out('output_parameters',
                 orm.Dict(dict=d_output_parameters).store())

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_workdir=True` in the inputs."""
        super().on_terminated()

        if self.inputs.clean_workdir.value is False:
            self.report('remote folders will not be cleaned')
            return

        cleaned_calcs = []

        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, orm.CalcJobNode):
                try:
                    called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                    cleaned_calcs.append(called_descendant.pk)
                except (IOError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )



# -*- coding: utf-8 -*-
"""Module with launch utitlies for the CLI."""
import os

import click


def launch_process(process, daemon, **inputs):
    """Launch a process with the given inputs.
    If not sent to the daemon, the results will be displayed after the calculation finishes.
    :param process: the process class
    :param daemon: boolean, if True will submit to the daemon instead of running in current interpreter
    :param inputs: inputs for the process
    """
    from aiida.engine import launch, Process, ProcessBuilder

    if isinstance(process, ProcessBuilder):
        process_name = process.process_class.__name__
    elif issubclass(process, Process):
        process_name = process.__name__
    else:
        raise TypeError(f'invalid type for process: {process}')

    if daemon:
        node = launch.submit(process, **inputs)
        click.echo(f'Submitted {process_name}<{node.pk}> to the daemon')
    else:
        if inputs.get('metadata', {}).get('dry_run', False):
            click.echo(f'Running a dry run for {process_name}...')
        else:
            click.echo(f'Running a {process_name}...')
        _, node = launch.run_get_node(process, **inputs)
        echo_process_results(node)


def echo_process_results(node):
    """Display a formatted table of the outputs registered for the given process node.
    :param node: the `ProcessNode` of a terminated process
    """
    from aiida.common.links import LinkType

    class_name = node.process_class.__name__
    outputs = node.get_outgoing(link_type=(LinkType.CREATE,
                                           LinkType.RETURN)).all()

    if hasattr(node, 'dry_run_info'):
        # It is a dry-run: get the information and print it
        rel_path = os.path.relpath(node.dry_run_info['folder'])
        click.echo(f"-> Files created in folder '{rel_path}'")
        click.echo(
            f"-> Submission script filename: '{node.dry_run_info['script_filename']}'"
        )
        return

    if node.is_finished and node.exit_message:
        state = f'{node.process_state.value} [{node.exit_status}] `{node.exit_message}`'
    elif node.is_finished:
        state = f'{node.process_state.value} [{node.exit_status}]'
    else:
        state = node.process_state.value

    click.echo(f'{class_name}<{node.pk}> terminated with state: {state}')

    if not outputs:
        click.echo(f'{class_name}<{node.pk}> registered no outputs')
        return

    click.echo(f"\n{'Output link':25s} Node pk and type")
    click.echo(f"{'-' * 60}")

    for triple in sorted(outputs, key=lambda triple: triple.link_label):
        click.echo(
            f'{triple.link_label:25s} {triple.node.__class__.__name__}<{triple.node.pk}> '
        )



# -*- coding: utf-8 -*-
"""
Birch-Murnaghan fit as calcfunction refactor from eosfit.py v3.1 write by Kurt Lejaeghere
Copyright (C) 2012 Kurt Lejaeghere <Kurt.Lejaeghere@UGent.be>, Center for
Molecular Modeling (CMM), Ghent University, Ghent, Belgium
"""
from aiida.engine import calcfunction, ExitCode
from aiida import orm


@calcfunction
def birch_murnaghan_fit(volume_energy: orm.Dict):
    """
    doc
    """
    # pylint: disable=invalid-name
    import numpy as np

    volumes = np.array(list(volume_energy['volumes'].values()))
    energies = np.array(list(volume_energy['energies'].values()))
    fitdata = np.polyfit(volumes**(-2. / 3.), energies, 3, full=True)

    ssr = fitdata[1]
    sst = np.sum((energies - np.average(energies))**2.)
    residuals0 = ssr / sst
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)

    volume0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            volume0 = x**(-3. / 2.)
            break

    if volume0 == 0:
        return ExitCode(100, f'get spurious volume0={volume0}.')

    if not isinstance(volume0, float):
        # In case where the fitting failed
        return ExitCode(101, f'get spurious volume0={volume0}')

    derivV2 = 4. / 9. * x**5. * deriv2(x)
    derivV3 = (-20. / 9. * x**(13. / 2.) * deriv2(x) -
               8. / 27. * x**(15. / 2.) * deriv3(x))
    bulk_modulus0 = derivV2 / x**(3. / 2.)
    bulk_deriv0 = -1 - x**(-3. / 2.) * derivV3 / derivV2
    energy0 = deriv0(volume0**(-2. / 3.))

    echarge = 1.60217733e-19
    return orm.Dict(
        dict={
            'volume0': round(volume0, 7),
            'energy0': round(energy0, 7),
            'bulk_modulus0': round(bulk_modulus0 * echarge * 1.0e21, 7),
            'bulk_deriv0': round(bulk_deriv0, 7),
            'residuals0': round(residuals0[0]),
            'volume0_unit': 'A^3/atom',
            'bulk_modulus0_unit': 'GPa',
        })



# -*- coding: utf-8 -*-
"""
Convergence workchain based on aiida-optimize by
inherit `aiida_optimize.engines._convergence::_ConvergenceImpl`
"""
import typing as ty
import itertools

import numpy as np

from aiida_optimize.engines.base import OptimizationEngineWrapper
from aiida_optimize.engines._convergence import _ConvergenceImpl
from aiida_optimize.engines._result_mapping import Result
from aiida_optimize.helpers import get_nested_result

from aiida import orm
from aiida.orm.nodes.data.base import to_aiida_type

__all__ = ['TwoInputsTwoFactorsConvergence']


class _TwoInputsTwoFactorsConvergenceImpl(_ConvergenceImpl):
    def __init__(
        self,
        *,
        input_values: ty.List[ty.Any],
        tol: float,
        conv_thr: float,
        input_key: str,
        extra_input_key: str,
        result_key: str,
        convergence_window: int,
        array_name: ty.Optional[str],
        current_index: int,
        result_values: ty.List[ty.Any],
        initialized: bool,
        logger: ty.Optional[ty.Any],
        result_state: ty.Optional[ty.Dict[int, Result]] = None,
    ):
        super().__init__(
            input_values=input_values,
            tol=tol,
            input_key=input_key,
            result_key=result_key,
            convergence_window=convergence_window,
            array_name=array_name,
            current_index=current_index,
            result_values=result_values,
            initialized=initialized,
            logger=logger,
            result_state=result_state,
        )

        self.extra_input_key = extra_input_key
        self.conv_thr = conv_thr

    @property
    def _num_new_iters(self) -> int:
        """
        Determine the minimum number of additional outputs to have a hope
        of converging in the next step.
        """
        distance_triangle = self._distance_triangle
        results_in_window = self._result_window
        # Find location of the last calculation which creates out-of-tolerance
        # roughness, and do enough calculations so that it is no longer in the
        # next convergence window
        num_new_iters_tol = 0
        for i, row in enumerate(distance_triangle):
            if np.any(np.array(row) > self.tol):
                num_new_iters_tol = i + 1

        # count the number of results that not satisfy conv_thr
        # the more not satisfied the more add to the next iteration
        # if all under the conv_thr add (zero) new iteration
        num_new_iters_conv_thr = (np.array(results_in_window) >
                                  self.conv_thr).sum()

        # Use the max(greedy) number of the two types of criterion
        num_new_iters = max(num_new_iters_tol, num_new_iters_conv_thr)

        # Check that we don't go past the end of the input_values when trying
        # to remove the calculation that is too rough from the window
        # If we do, return -1 as an indication that convergence will not be
        # possible
        if self.current_index + num_new_iters > len(self.input_values):
            # num_new_iters = len(self.input_values) - self.current_index
            num_new_iters = -1

        return num_new_iters

    @property
    def is_converged(self) -> bool:
        """
        Check if convergence has been reached by calculating the Frobenius
        or 2-norm of the difference between all the result values / arrays
        and checking that the maximum distance between points in the
        convergence window is less than the tolerance.
        """
        if not self.initialized:
            return False

        # calculate pair distances between results
        distance_triangle = self._distance_triangle
        # flatten all the distances into a 1D list
        distances = list(itertools.chain(*distance_triangle))

        # check if the maximum distance is less than the tolerance
        # import ipdb; ipdb.set_trace()
        is_tol_converge = bool(np.max(distances) < self.tol)
        is_conv_thr_converge = bool(
            np.max(self._result_window) < self.conv_thr)

        res = is_tol_converge and is_conv_thr_converge

        return res

    def _create_inputs(self) -> ty.List[ty.Dict[str, orm.Float]]:
        """
        Create the inputs for the evaluation function.
        If the work chain is not initialized, the appropriate number of
        inputs to fill the convergence window are generated.
        Otherwise, one more input will be generated
        """
        if not self.initialized:
            # Do enough calculations to fill the initial convergence window
            num_new_iters = self.convergence_window
            self.initialized = True
        else:
            num_new_iters = self._num_new_iters

        self.current_index += num_new_iters
        inputs = [{
            self.input_key: to_aiida_type(self.input_values[i][0]),
            self.extra_input_key: to_aiida_type(self.input_values[i][1]),
        } for i in range(self.current_index -
                         num_new_iters, self.current_index)]

        return inputs

    def _get_optimal_result(self) -> ty.Tuple[int, orm.Node, orm.Node]:
        """
        Retrieve the converged index and result value (output value, _not_ max
        distance within convergence window)
        """
        opt_index = len(self.result_values) - self.convergence_window
        opt_input = self._result_mapping[opt_index].input[self.input_key]
        opt_output = get_nested_result(self._result_mapping[opt_index].output,
                                       self.result_key)

        return opt_index, opt_input, opt_output


class TwoInputsTwoFactorsConvergence(OptimizationEngineWrapper):
    """
    Wrapper class for convergence engine
    """

    _IMPL_CLASS = _TwoInputsTwoFactorsConvergenceImpl

    def __new__(  # pylint: disable=too-many-arguments,arguments-differ
        cls,
        input_values: ty.List[ty.Any],
        tol: float,
        conv_thr: float,
        input_key: str,
        extra_input_key: str,
        result_key: str,
        convergence_window: int = 2,
        array_name: ty.Optional[str] = None,
        logger: ty.Optional[ty.Any] = None,
    ) -> _TwoInputsTwoFactorsConvergenceImpl:
        return cls._IMPL_CLASS(  # pylint: disable=no-member
            input_values=input_values,
            tol=tol,
            conv_thr=conv_thr,
            input_key=input_key,
            extra_input_key=extra_input_key,
            result_key=result_key,
            convergence_window=convergence_window,
            array_name=array_name,
            current_index=0,
            result_values=[],
            initialized=False,
            logger=logger)



# -*- coding: utf-8 -*-
"""
Module comtain helper functions for workflows
"""
import importlib_resources

from aiida import orm
from aiida.engine import calcfunction
from aiida.plugins import DataFactory
from aiida.tools.data.structure import spglib_tuple_to_structure, structure_to_spglib_tuple
import seekpath

from aiida_sssp_workflow.utils import RARE_EARTH_ELEMENTS, \
    get_standard_cif_filename_from_element

UpfData = DataFactory('pseudo.upf')


@calcfunction
def helper_get_primitive_structure(structure,
                                   **parameters) -> orm.StructureData:
    """
    :param structure: The AiiDA StructureData for which we want to obtain
        the primitive structure.

    :param parameters: A dictionary whose key-value pairs are passed as
        additional kwargs to the ``seekpath.get_explicit_k_path`` function.

    """
    structure_tuple, kind_info, kinds = structure_to_spglib_tuple(structure)

    rawdict = seekpath.get_explicit_k_path(structure=structure_tuple,
                                           **parameters)

    # Replace primitive structure with AiiDA StructureData
    primitive_lattice = rawdict.pop('primitive_lattice')
    primitive_positions = rawdict.pop('primitive_positions')
    primitive_types = rawdict.pop('primitive_types')
    primitive_tuple = (primitive_lattice, primitive_positions, primitive_types)
    primitive_structure = spglib_tuple_to_structure(primitive_tuple, kind_info,
                                                    kinds)

    return primitive_structure


def helper_get_base_inputs(pseudo: UpfData, primitive_cell=True):
    """
    helper method used to generate base pw inputs(structure, pseudos, pw_parameters).
    lanthanides elements are supported with Rare-Nithides.
    """
    element = pseudo.element

    pseudos = {element: pseudo}

    if element == 'F':
        # set element to 'SiF4' to use SiF4 structure for fluorine
        element = orm.Str('SiF4')

        fpath = importlib_resources.path('aiida_sssp_workflow.REF.UPFs',
                                         'Si.pbe-n-rrkjus_psl.1.0.0.UPF')
        with fpath as path:
            filename = str(path)
            upf_silicon = UpfData.get_or_create(filename)[0]
            pseudos['Si'] = upf_silicon

    cif_file = get_standard_cif_filename_from_element(element)

    cif_data = orm.CifData.get_or_create(cif_file)[0]

    return {
        'structure': cif_data.get_structure(primitive_cell=primitive_cell),
        'pseudos': pseudos,
        'base_pw_parameters': pw_parameters,
    }


def helper_get_v0_b0_b1(element: str):
    """get eos reference of element"""
    import re
    from aiida_sssp_workflow.calculations.wien2k_ref import WIEN2K_REF, WIEN2K_REN_REF

    if element == 'F':
        # Use SiF4 as reference of fluorine(F)
        return 19.3583, 74.0411, 4.1599

    if element in RARE_EARTH_ELEMENTS:
        element_str = f'{element}N'
    else:
        element_str = element

    regex = re.compile(
        rf"""{element_str}\s*
                        (?P<V0>\d*.\d*)\s*
                        (?P<B0>\d*.\d*)\s*
                        (?P<B1>\d*.\d*)""", re.VERBOSE)
    if element not in RARE_EARTH_ELEMENTS:
        match = regex.search(WIEN2K_REF)
        V0 = match.group('V0')
        B0 = match.group('B0')
        B1 = match.group('B1')
    else:
        match = regex.search(WIEN2K_REN_REF)
        V0 = match.group('V0')
        B0 = match.group('B0')
        B1 = match.group('B1')

    return float(V0), float(B0), float(B1)



# -*- coding: utf-8 -*-
"""utils.py"""

import collections.abc
import importlib_resources

from aiida.plugins import DataFactory

UpfData = DataFactory('pseudo.upf')

RARE_EARTH_ELEMENTS = [
    'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
    'Tm', 'Yb', 'Lu'
]

MAGNETIC_ELEMENTS = ['Mn', 'O', 'Cr', 'Fe', 'Co', 'Ni']

NONMETAL_ELEMENTS = [
    'H', 'He', 'B', 'N', 'O', 'F', 'Ne', 'Si', 'P', 'Cl', 'Ar', 'Se', 'Br',
    'Kr', 'Te', 'I', 'Xe', 'Rn'
]


def update_dict(d, u):
    """update dict by7 another dict, update include all hierarchy
    return the update dict instead of change the input dict"""
    # pylint: disable=invalid-name
    import copy

    ret = copy.deepcopy(d)
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            ret[k] = update_dict(ret.get(k, {}), v)
        else:
            ret[k] = v
    return ret


def get_standard_cif_filename_from_element(element: str) -> str:
    """
    get cif filename from element

    NOTICE!!: that for convergence verification the SiF4 structure is used, the
    name of file is `SiF4.cif` which can be get by `element=SiF4`.
    """
    if element in RARE_EARTH_ELEMENTS:
        fpath = importlib_resources.path('aiida_sssp_workflow.REF.CIFs_REN',
                                         f'{element}N.cif')
    else:
        fpath = importlib_resources.path('aiida_sssp_workflow.REF.CIFs',
                                         f'{element}.cif')
    with fpath as path:
        filename = str(path)

    return filename


def parse_upf(upf_content: str) -> dict:
    """

    :param upf_content:
    :param check: if check the integrity of the pp file
    :return:
    """
    from upf_to_json import upf_to_json

    upf_dict = upf_to_json(upf_content, None)

    return upf_dict['pseudo_potential']


def helper_parse_upf(upf: UpfData) -> dict:
    """parser upf"""
    header = parse_upf(upf.get_content())['header']

    return header


def get_default_options(max_num_machines=1,
                        max_wallclock_seconds=1800,
                        with_mpi=False):
    """Return an instance of the options dictionary with the minimally required parameters for a `CalcJob`.

    :param max_num_machines: set the number of nodes, default=1
    :param max_wallclock_seconds: set the maximum number of wallclock seconds, default=1800
    :param with_mpi: whether to run the calculation with MPI enabled
    """
    return {
        'resources': {
            'num_machines': int(max_num_machines)
        },
        'max_wallclock_seconds': int(max_wallclock_seconds),
        'withmpi': with_mpi,
    }


def to_valid_key(name):
    """
    convert name into a valid key name which contain only alphanumeric and underscores
    """
    import re

    valid_name = re.sub(r'[^\w\s]', '_', name)

    return valid_name



# -*- coding: utf-8 -*-
"""
Refactor from calcDelta.py v3.1 write by Kurt Lejaeghere
Copyright (C) 2012 Kurt Lejaeghere <Kurt.Lejaeghere@UGent.be>, Center for
Molecular Modeling (CMM), Ghent University, Ghent, Belgium
"""
from io import StringIO

import numpy as np
from aiida.engine import calcfunction
from aiida import orm

from aiida_sssp_workflow.calculations.wien2k_ref import WIEN2K_REF, WIEN2K_REN_REF
from aiida_sssp_workflow.helpers import helper_get_v0_b0_b1

# pylint: disable=invalid-name

RARE_EARTH_ELEMENTS = [
    'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
    'Tm', 'Yb', 'Lu'
]


@calcfunction
def calculate_delta(element, V0, B0, B1) -> orm.Dict:
    """
    The calcfunction calculate the delta factor.
    return delta factor with unit (eV/atom)
    """
    if element.value in RARE_EARTH_ELEMENTS:
        wien2k_ref = StringIO(WIEN2K_REN_REF)
    else:
        # not Lanthanides elements
        wien2k_ref = StringIO(WIEN2K_REF)

    data_ref = np.loadtxt(wien2k_ref,
                          dtype={
                              'names': ('element', 'V0', 'B0', 'B1'),
                              'formats': ('U2', np.float, np.float, np.float)
                          })
    ref_V0, ref_B0, ref_B1 = helper_get_v0_b0_b1(element.value)

    # Here use dtype U2 to truncate REN to RE
    data_tested = np.array(
        [(element.value, V0.value, B0.value, B1.value)],
        dtype={
            'names': ('element', 'V0', 'B0', 'B1'),
            'formats': ('U2', np.float, np.float, np.float),
        })

    eloverlap = [element.value]
    if not eloverlap:
        # TODO ExitCode
        raise ValueError(
            f'Element {element} is not present in the reference set')
    eloverlap = [element.value]
    # Delta computation
    Delta, Deltarel, Delta1 = _calcDelta(data_tested, data_ref, eloverlap)

    return orm.Dict(
        dict={
            'delta': Delta[0],
            'delta1': Delta1[0],
            'delta_unit': 'meV/atom',
            'delta_relative': Deltarel[0],
            'delta_relative_unit': '%',
            'birch_murnaghan_results': [V0, B0, B1],
            'reference_wien2k_V0_B0_B1': [ref_V0, ref_B0, ref_B1]
        })


def _calcDelta(data_f, data_w, eloverlap, useasymm=False):
    """
    Calculate the Delta using the data in data_f, data_w on
    element in eloverlap
    data_w is all-electrons result as ref
    """
    # pylint: disable=too-many-statements, consider-using-enumerate
    v0w = np.zeros(len(eloverlap))
    b0w = np.zeros(len(eloverlap))
    b1w = np.zeros(len(eloverlap))

    v0f = np.zeros(len(eloverlap))
    b0f = np.zeros(len(eloverlap))
    b1f = np.zeros(len(eloverlap))

    elw = list(data_w['element'])
    elf = list(data_f['element'])

    for i in range(len(eloverlap)):
        searchnr = elw.index(eloverlap[i])
        v0w[i] = data_w['V0'][searchnr]
        b0w[i] = data_w['B0'][searchnr] * 10.**9. / 1.602176565e-19 / 10.**30.
        b1w[i] = data_w['B1'][searchnr]

        searchnr = elf.index(eloverlap[i])
        v0f[i] = data_f['V0'][searchnr]
        b0f[i] = data_f['B0'][searchnr] * 10.**9. / 1.602176565e-19 / 10.**30.
        b1f[i] = data_f['B1'][searchnr]

    vref = 30.
    bref = 100. * 10.**9. / 1.602176565e-19 / 10.**30.

    if useasymm:
        Vi = 0.94 * v0w
        Vf = 1.06 * v0w
    else:
        Vi = 0.94 * (v0w + v0f) / 2.
        Vf = 1.06 * (v0w + v0f) / 2.

    a3f = 9. * v0f**3. * b0f / 16. * (b1f - 4.)
    a2f = 9. * v0f**(7. / 3.) * b0f / 16. * (14. - 3. * b1f)
    a1f = 9. * v0f**(5. / 3.) * b0f / 16. * (3. * b1f - 16.)
    a0f = 9. * v0f * b0f / 16. * (6. - b1f)

    a3w = 9. * v0w**3. * b0w / 16. * (b1w - 4.)
    a2w = 9. * v0w**(7. / 3.) * b0w / 16. * (14. - 3. * b1w)
    a1w = 9. * v0w**(5. / 3.) * b0w / 16. * (3. * b1w - 16.)
    a0w = 9. * v0w * b0w / 16. * (6. - b1w)

    x = [0, 0, 0, 0, 0, 0, 0]

    x[0] = (a0f - a0w)**2
    x[1] = 6. * (a1f - a1w) * (a0f - a0w)
    x[2] = -3. * (2. * (a2f - a2w) * (a0f - a0w) + (a1f - a1w)**2.)
    x[3] = -2. * (a3f - a3w) * (a0f - a0w) - 2. * (a2f - a2w) * (a1f - a1w)
    x[4] = -3. / 5. * (2. * (a3f - a3w) * (a1f - a1w) + (a2f - a2w)**2.)
    x[5] = -6. / 7. * (a3f - a3w) * (a2f - a2w)
    x[6] = -1. / 3. * (a3f - a3w)**2.

    y = [0, 0, 0, 0, 0, 0, 0]

    y[0] = (a0f + a0w)**2 / 4.
    y[1] = 3. * (a1f + a1w) * (a0f + a0w) / 2.
    y[2] = -3. * (2. * (a2f + a2w) * (a0f + a0w) + (a1f + a1w)**2.) / 4.
    y[3] = -(a3f + a3w) * (a0f + a0w) / 2. - (a2f + a2w) * (a1f + a1w) / 2.
    y[4] = -3. / 20. * (2. * (a3f + a3w) * (a1f + a1w) + (a2f + a2w)**2.)
    y[5] = -3. / 14. * (a3f + a3w) * (a2f + a2w)
    y[6] = -1. / 12. * (a3f + a3w)**2.

    Fi = np.zeros_like(Vi)
    Ff = np.zeros_like(Vf)

    Gi = np.zeros_like(Vi)
    Gf = np.zeros_like(Vf)

    for n in range(7):
        Fi = Fi + x[n] * Vi**(-(2. * n - 3.) / 3.)
        Ff = Ff + x[n] * Vf**(-(2. * n - 3.) / 3.)

        Gi = Gi + y[n] * Vi**(-(2. * n - 3.) / 3.)
        Gf = Gf + y[n] * Vf**(-(2. * n - 3.) / 3.)

    Delta = 1000. * np.sqrt((Ff - Fi) / (Vf - Vi))
    Deltarel = 100. * np.sqrt((Ff - Fi) / (Gf - Gi))
    if useasymm:
        Delta1 = 1000. * np.sqrt((Ff - Fi) / (Vf - Vi)) \
                 / v0w / b0w * vref * bref
    else:
        Delta1 = 1000. * np.sqrt((Ff - Fi) / (Vf - Vi)) \
                 / (v0w + v0f) / (b0w + b0f) * 4. * vref * bref

    return Delta, Deltarel, Delta1



# -*- coding: utf-8 -*-
"""
Convergence test on cohesive energy of a given pseudopotential
"""
from aiida.engine import calcfunction
from aiida import orm

from aiida_sssp_workflow.utils import update_dict
from aiida_sssp_workflow.workflows.cohesive_energy import CohesiveEnergyWorkChain
from .base import BaseConvergenceWorkChain


@calcfunction
def helper_cohesive_energy_difference(input_parameters: orm.Dict,
                                      ref_parameters: orm.Dict) -> orm.Dict:
    """calculate the cohesive energy difference from parameters"""
    res_energy = input_parameters['cohesive_energy_per_atom']
    ref_energy = ref_parameters['cohesive_energy_per_atom']
    absolute_diff = abs(res_energy - ref_energy)
    relative_diff = abs((res_energy - ref_energy) / ref_energy) * 100

    res = {
        'absolute_diff': absolute_diff,
        'relative_diff': relative_diff,
        'absolute_unit': 'eV/atom',
        'relative_unit': '%'
    }

    return orm.Dict(dict=res)


PARA_ECUTWFC_LIST = lambda: orm.List(list=[
    20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 110,
    120, 130, 140, 160, 180, 200
])

PARA_ECUTRHO_LIST = lambda: orm.List(list=[
    160, 200, 240, 280, 320, 360, 400, 440, 480, 520, 560, 600, 640, 680, 720,
    760, 800, 880, 960, 1040, 1120, 1280, 1440, 1600
])


class ConvergenceCohesiveEnergyWorkChain(BaseConvergenceWorkChain):
    """WorkChain to converge test on cohisive energy of input structure"""
    # pylint: disable=too-many-instance-attributes

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('code',
                   valid_type=orm.Code,
                   help='The `pw.x` code use for the `PwCalculation`.')

    def setup_protocol(self):
        # pylint: disable=invalid-name, attribute-defined-outside-init
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['convergence']['cohesive_energy']
        self.ctx._DEGAUSS = protocol['degauss']
        self.ctx._OCCUPATIONS = protocol['occupations']
        self.ctx._BULK_SMEARING = protocol['bulk_smearing']
        self.ctx._ATOM_SMEARING = protocol['atom_smearing']
        self.ctx._CONV_THR_EVA = protocol['electron_conv_thr']
        self.ctx._KDISTANCE = protocol['kpoints_distance']
        self.ctx._VACUUM_LENGTH = protocol['vaccum_length']

        self.ctx._TOLERANCE = protocol['tolerance']
        self.ctx._CONV_THR_CONV = protocol['convergence_conv_thr']
        self.ctx._CONV_WINDOW = protocol['convergence_window']

    def get_create_process(self):
        return CohesiveEnergyWorkChain

    def get_evaluate_process(self):
        return helper_cohesive_energy_difference

    def get_parsed_results(self):
        return {
            'absolute_diff': ('The absolute cohesive difference', 'eV/atom'),
            'relative_diff': ('The relative cohesive difference', '%'),
        }

    def get_converge_y(self):
        return 'absolute_diff', 'eV/atom'

    def get_create_process_inputs(self):
        _PW_BULK_PARAS = {   # pylint: disable=invalid-name
            'SYSTEM': {
                'degauss': self.ctx._DEGAUSS,
                'occupations': self.ctx._OCCUPATIONS,
                'smearing': self.ctx._BULK_SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self.ctx._CONV_THR_EVA,
            },
        }
        _PW_ATOM_PARAS = {   # pylint: disable=invalid-name
            'SYSTEM': {
                'degauss': self.ctx._DEGAUSS,
                'occupations': self.ctx._OCCUPATIONS,
                'smearing': self.ctx._ATOM_SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self.ctx._CONV_THR_EVA,
            },
        }
        inputs = {
            'code': self.inputs.code,
            'pseudos': self.ctx.pseudos,
            'structure': self.ctx.structure,
            'parameters': {
                'pw_bulk':
                orm.Dict(dict=update_dict(_PW_BULK_PARAS,
                                          self.ctx.base_pw_parameters)),
                'pw_atom':
                orm.Dict(dict=_PW_ATOM_PARAS),
                'kpoints_distance':
                orm.Float(self.ctx._KDISTANCE),
                'vacuum_length':
                orm.Float(self.ctx._VACUUM_LENGTH),
            },
        }

        return inputs

    def get_evaluate_process_inputs(self):
        ref_workchain = self.ctx.ref_workchain

        res = {
            'ref_parameters': ref_workchain.outputs.output_parameters,
        }

        return res

    def get_output_input_mapping(self):
        res = orm.Dict(dict={'output_parameters': 'input_parameters'})
        return res



# -*- coding: utf-8 -*-
"""Equation of state workflow that can use any code plugin implementing the common relax workflow."""
from aiida import orm
from aiida.common import AttributeDict
from aiida.engine import WorkChain, append_, calcfunction
from aiida.plugins import WorkflowFactory, CalculationFactory

PwBaseWorkChain = WorkflowFactory('quantumespresso.pw.base')
birch_murnaghan_fit = CalculationFactory('sssp_workflow.birch_murnaghan_fit')


def validate_inputs(value, _):
    """Validate the entire input namespace."""
    if 'scale_factors' not in value and ('scale_count' not in value
                                         and 'scale_count' not in value):
        return 'neither `scale_factors` nor the pair of `scale_count` and `scale_increment` were defined.'


def validate_scale_factors(value, _):
    """Validate the `validate_scale_factors` input."""
    if value and len(value) < 3:
        return 'need at least 3 scaling factors.'


def validate_scale_count(value, _):
    """Validate the `scale_count` input."""
    if value is not None and value < 3:
        return 'need at least 3 scaling factors.'


def validate_scale_increment(value, _):
    """Validate the `scale_increment` input."""
    if value is not None and not 0 < value < 1:
        return 'scale increment needs to be between 0 and 1.'


@calcfunction
def scale_structure(structure: orm.StructureData,
                    scale_factor: orm.Float) -> orm.StructureData:
    """Scale the structure with the given scaling factor."""
    ase = structure.get_ase().copy()
    ase.set_cell(ase.get_cell() * float(scale_factor)**(1 / 3),
                 scale_atoms=True)
    return orm.StructureData(ase=ase)


class _EquationOfStateWorkChain(WorkChain):
    """Workflow to compute the equation of state for a given crystal structure."""
    @classmethod
    def define(cls, spec):
        # yapf: disable
        super().define(spec)
        spec.expose_inputs(PwBaseWorkChain, namespace='scf',
            exclude=('clean_workdir', 'pw.structure', 'pw.kpoints', 'pw.kpoints_distance'),
            namespace_options={'help': 'Inputs for the `PwBaseWorkChain` for the SCF calculation.'})
        spec.input('structure', valid_type=orm.StructureData, help='The structure at equilibrium volume.')
        spec.input('kpoints_distance', valid_type=orm.Float, required=True,
                   help='The kpoints distance used in generating the kmesh of '
                        'unscaled structure then for all scaled structures')
        spec.input('scale_factors', valid_type=orm.List, required=False, validator=validate_scale_factors,
            help='The list of scale factors at which the volume and total energy of the structure should be computed.')
        spec.input('scale_count', valid_type=orm.Int, default=lambda: orm.Int(7), validator=validate_scale_count,
            help='The number of points to compute for the equation of state.')
        spec.input('scale_increment', valid_type=orm.Float, default=lambda: orm.Float(0.02),
            validator=validate_scale_increment,
            help='The relative difference between consecutive scaling factors.')
        spec.inputs.validator = validate_inputs
        spec.outline(
            cls.run_init,
            cls.run_eos,
            cls.inspect_eos,
        )
        spec.output('output_volume_energy', valid_type=orm.Dict,
                    help='Results volumes and energise.')
        spec.output('output_birch_murnaghan_fit', valid_type=orm.Dict,
                    help='Result of birch murnaghan fitting.')
        spec.exit_code(400, 'ERROR_SUB_PROCESS_FAILED',
            message='At least one of the `{cls}` sub processes did not finish successfully.')
        # yapf: enable

    def get_scale_factors(self):
        """Return the list of scale factors."""
        if 'scale_factors' in self.inputs:
            return self.inputs.scale_factors

        count = self.inputs.scale_count.value
        increment = self.inputs.scale_increment.value
        return [
            orm.Float(1 + i * increment - (count - 1) * increment / 2)
            for i in range(count)
        ]

    def get_sub_workchain_builder(self, scale_factor):
        """Return the builder for the relax workchain."""
        process_class = PwBaseWorkChain
        structure = scale_structure(self.inputs.structure, scale_factor)
        unscaled_structure = self.inputs.structure

        inputs = AttributeDict(
            self.exposed_inputs(PwBaseWorkChain, namespace='scf'))

        parameters = inputs.pw.parameters.get_dict()
        parameters.setdefault('CONTROL', {})['calculation'] = 'scf'

        kpoints = orm.KpointsData()
        kpoints.set_cell_from_structure(unscaled_structure)
        kpoints.set_kpoints_mesh_from_density(
            distance=self.inputs.kpoints_distance.value)

        inputs.metadata.call_link_label = 'scf'
        inputs.kpoints = kpoints.store()
        inputs.pw.structure = structure
        inputs.pw.parameters = orm.Dict(dict=parameters)
        if 'options' in self.inputs:
            inputs.pw.metadata.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            inputs.pw.metadata.options = get_default_options(with_mpi=True)

        builder = process_class.get_builder()

        builder.update(**inputs)

        return builder

    def run_init(self):
        """Run the first sub-workchain, if this failed the whole workchain break."""
        scale_factor = self.get_scale_factors()[0]
        builder = self.get_sub_workchain_builder(scale_factor)
        self.report(
            f'submitting `{builder.process_class.__name__}` for scale_factor `{scale_factor}`'
        )
        self.ctx.previous_workchain = self.submit(builder)
        self.to_context(children=append_(self.ctx.previous_workchain))

    def run_eos(self):
        """Run the sub process at each scale factor to compute the structure volume and total energy."""
        workchain = self.ctx.children[0]

        if not workchain.is_finished_ok:
            self.report(
                f'PwBaseWorkChain pk={workchain.pk} for first scale structure run is failed.'
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                cls=PwBaseWorkChain)

        for scale_factor in self.get_scale_factors()[1:]:
            builder = self.get_sub_workchain_builder(scale_factor)
            self.report(
                f'submitting `{builder.process_class.__name__}` for scale_factor `{scale_factor}`'
            )
            self.to_context(children=append_(self.submit(builder)))

    def inspect_eos(self):
        """Inspect all children workflows to make sure they finished successfully."""
        if any(not child.is_finished_ok for child in self.ctx.children):
            process_class = PwBaseWorkChain
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                cls=process_class)

        volume_energy = {
            'volumes': {},
            'energies': {},
            'volume_unit': 'A^3/atom',
            'energy_unit': 'eV/atom',
        }
        for index, child in enumerate(self.ctx.children):
            volume = child.outputs.output_parameters['volume']
            energy = child.outputs.output_parameters[
                'energy']  # Already the free energy E-TS (metal)
            num_of_atoms = child.outputs.output_parameters['number_of_atoms']
            self.report(
                f'Image {index}: volume={volume}, total energy={energy}')
            volume_energy['volumes'][index] = volume / num_of_atoms
            volume_energy['energies'][index] = energy / num_of_atoms

        output_volume_energy = orm.Dict(dict=volume_energy).store()
        output_birch_murnaghan_fit = birch_murnaghan_fit(output_volume_energy)

        self.report(
            f'The birch murnaghan fitting results are: {output_birch_murnaghan_fit.get_dict()}'
        )

        self.out('output_volume_energy', output_volume_energy)
        self.out('output_birch_murnaghan_fit', output_birch_murnaghan_fit)
